#!/usr/bin/env python
"""
Unit tests for the AuthzByStatus module
"""
import unittest

from WMCore.ReqMgr.DataStructs.AuthzByStatus import AuthzByStatus

ADMIN = {'role': ["admin_role"], 'group': ["admin_group"]}
OPS = {'role': ["admin_role", "ops_role"], 'group': ["admin_group", "ops_group"]}
PPD = {'role': ["admin_role", "ops_role", "ppd_role"],
       'group': ["admin_group", "ops_group", "ppd_group"]}


class AuthzByStatusTests(unittest.TestCase):
    """Some basic tests for AuthzByStatus"""

    def setUp(self):
        self.authorized_roles = {"admin": ADMIN,
                                 "ops": OPS,
                                 "ppd": PPD}
        self.authz_by_status = [{"permission": "admin",
                                 "statuses": ["acquired", "running-open", "running-closed", "completed",
                                              "aborted-completed", "failed", "rejected-archived",
                                              "aborted-archived", "normal-archived"]},
                                {"permission": "ops",
                                 "statuses": ["assigned", "staging", "staged", "force-complete",
                                              "closed-out", "announced"]},
                                {"permission": "ppd",
                                 "statuses": ["new", "assignment-approved", "rejected", "aborted", "NO_STATUS"]}]

        # self.assertItemsEqual = self.assertCountEqual

    def tearDown(self):
        pass

    def testNoStatus(self):
        """Test case where NO_STATUS is not provided"""
        with self.assertRaises(RuntimeError):
            AuthzByStatus([], {})

    def testBadPermissions(self):
        """Test case for lack of group permissions"""
        with self.assertRaises(RuntimeError):
            AuthzByStatus([], {"admin": "a", "ops": "o"})

    def testBadStatusValue(self):
        """Test case for bad statuses-based permissions"""
        with self.assertRaises(RuntimeError):
            AuthzByStatus([{"permission": "ALAN", "statuses": ["new", "acquired"]}],
                          {"admin": "a", "ops": "o", "ppd": "e", "ALAN": "aaa"})

    def testBadStatusName(self):
        """Test case for an unexpected status name"""
        with self.assertRaises(RuntimeError):
            AuthzByStatus([{"permission": "admin", "statuses": ["new", "assigned"]},
                           {"permission": "ops", "statuses": ["staging", "staged"]},
                           {"permission": "ppd", "statuses": ["acquired", "Alan"]}],
                          {"admin": "a", "ops": "o", "ppd": "e"})

    def testCorrectConfig(self):
        """Test correct configuration for the authorization setup"""
        authzPerm = AuthzByStatus(self.authz_by_status, self.authorized_roles)

        # test no status permission
        resp = authzPerm.getRolesGroupsByStatus({})
        self.assertCountEqual(resp, PPD)

        # test admin permissions (and one allowed by Ops and PPD)
        for status in {"new", "failed", "normal-archived", "staged", "aborted"}:
            resp = authzPerm.getRolesGroupsByStatus({"RequestStatus": status})
            self.assertCountEqual(resp, ADMIN)

        # test Ops permissions (and one allowed by PPD)
        for status in {"assigned", "staged", "announced", "aborted"}:
            resp = authzPerm.getRolesGroupsByStatus({"RequestStatus": status})
            self.assertCountEqual(resp, OPS)

        # test PPD permissions
        for status in {"new", "assignment-approved", "rejected", "aborted"}:
            resp = authzPerm.getRolesGroupsByStatus({"RequestStatus": status})
            self.assertCountEqual(resp, OPS)



#!/usr/bin/env python
"""
_JobSubmitterCaching_t_

Verify that the caching of jobs and white/black lists works correctly.
"""

import os
import pickle
import unittest

from WMComponent.JobSubmitter.JobSubmitterPoller import JobSubmitterPoller
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMBase import getTestBase
from WMCore.WorkQueue.WMBSHelper import killWorkflow
from WMQuality.Emulators import EmulatorSetup
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase


class JobSubmitterCachingTest(EmulatedUnitTestCase):
    def setUp(self):
        """
        _setUp_

        Set everything up.
        """

        super(JobSubmitterCachingTest, self).setUp()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS", "WMCore.BossAir",
                                               "WMCore.ResourceControl"],
                                useDefault=False)
        self.testInit.setupCouch("jobsubmittercaching_t/jobs", "JobDump")
        self.testInit.setupCouch("jobsubmittercaching_t/fwjrs", "FWJRDump")

        resourceControl = ResourceControl()
        for siteName in ["T1_US_FNAL", "T1_UK_RAL"]:
            resourceControl.insertSite(siteName=siteName, pnn="%s_Disk" % (siteName),
                                       ceName=siteName, plugin="SimpleCondorPlugin", cmsName=siteName)
            resourceControl.insertThreshold(siteName=siteName, taskType="Processing",
                                            maxSlots=10000, pendingSlots=10000)

        self.testDir = self.testInit.generateWorkDir()
        self.configFile = EmulatorSetup.setupWMAgentConfig()
        return

    def tearDown(self):
        """
        _tearDown_

        Tear everything down.
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)
        return

    def createConfig(self):
        """
        _createConfig_

        Create a config for the JobSubmitter.  These parameters are still pulled
        from the environment.
        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        config.component_("Agent")
        config.Agent.isDocker = False

        config.section_("JobStateMachine")
        config.JobStateMachine.couchurl = os.getenv("COUCHURL")
        config.JobStateMachine.couchDBName = "jobsubmittercaching_t"

        config.section_("BossAir")
        config.BossAir.pluginDir = "WMCore.BossAir.Plugins"
        config.BossAir.pluginNames = ["SimpleCondorPlugin"]
        config.BossAir.nCondorProcesses = 1

        config.component_("JobSubmitter")
        config.JobSubmitter.submitDir = self.testDir
        config.JobSubmitter.submitScript = os.path.join(getTestBase(),
                                                        'WMComponent_t/JobSubmitter_t',
                                                        'submit.sh')

        return config

    def injectJobs(self):
        """
        _injectJobs_

        Inject two workflows into WMBS and save the job objects to disk.
        """
        testWorkflowA = Workflow(spec="specA.pkl", owner="Steve",
                                 name="wf001", task="TestTaskA")
        testWorkflowA.create()
        testWorkflowB = Workflow(spec="specB.pkl", owner="Steve",
                                 name="wf002", task="TestTaskB")
        testWorkflowB.create()

        testFileset = Fileset("testFileset")
        testFileset.create()

        testSubA = Subscription(fileset=testFileset, workflow=testWorkflowA)
        testSubA.create()
        testSubB = Subscription(fileset=testFileset, workflow=testWorkflowB)
        testSubB.create()

        testGroupA = JobGroup(subscription=testSubA)
        testGroupA.create()
        testGroupB = JobGroup(subscription=testSubB)
        testGroupB.create()

        stateChanger = ChangeState(self.createConfig(), "jobsubmittercaching_t")

        for i in range(10):
            newFile = File(lfn="testFile%s" % i,
                           locations=set(["se.T1_US_FNAL", "se.T1_UK_RAL"]))
            newFile.create()

            newJobA = Job(name="testJobA-%s" % i, files=[newFile])
            newJobA["workflow"] = "wf001"
            newJobA["possiblePSN"] = ["T1_US_FNAL"]
            newJobA["sandbox"] = "%s/somesandbox" % self.testDir
            newJobA["owner"] = "Steve"

            jobCacheDir = os.path.join(self.testDir, "jobA-%s" % i)
            os.mkdir(jobCacheDir)
            newJobA["cache_dir"] = jobCacheDir
            newJobA["type"] = "Processing"
            newJobA['requestType'] = 'ReReco'
            newJobA.create(testGroupA)

            jobHandle = open(os.path.join(jobCacheDir, "job.pkl"), "wb")
            pickle.dump(newJobA, jobHandle)
            jobHandle.close()

            stateChanger.propagate([newJobA], "created", "new")

            newJobB = Job(name="testJobB-%s" % i, files=[newFile])
            newJobB["workflow"] = "wf001"
            newJobB["possiblePSN"] = ["T1_UK_RAL"]
            newJobB["sandbox"] = "%s/somesandbox" % self.testDir
            newJobB["owner"] = "Steve"

            jobCacheDir = os.path.join(self.testDir, "jobB-%s" % i)
            os.mkdir(jobCacheDir)
            newJobB["cache_dir"] = jobCacheDir
            newJobB["type"] = "Processing"
            newJobB['requestType'] = 'ReReco'
            newJobB.create(testGroupB)

            jobHandle = open(os.path.join(jobCacheDir, "job.pkl"), "wb")
            pickle.dump(newJobB, jobHandle)
            jobHandle.close()

            stateChanger.propagate([newJobB], "created", "new")

        return

    def testCaching(self):
        """
        _testCaching_

        Verify that JobSubmitter caching works.
        """
        config = self.createConfig()
        mySubmitterPoller = JobSubmitterPoller(config)
        mySubmitterPoller.getThresholds()
        mySubmitterPoller.refreshCache()

        self.assertEqual(len(mySubmitterPoller.jobDataCache), 0,
                         "Error: The job cache should be empty.")

        self.injectJobs()
        mySubmitterPoller.refreshCache()

        # Verify the cache is full
        self.assertEqual(len(mySubmitterPoller.jobDataCache), 20,
                         "Error: The job cache should contain 20 jobs.  Contains: %i" % len(
                             mySubmitterPoller.jobDataCache))

        killWorkflow("wf001", jobCouchConfig=config)
        mySubmitterPoller.refreshCache()

        # Verify that the workflow is gone from the cache
        self.assertEqual(len(mySubmitterPoller.jobDataCache), 10,
                         "Error: The job cache should contain 10 jobs. Contains: %i" % len(
                             mySubmitterPoller.jobDataCache))

        killWorkflow("wf002", jobCouchConfig=config)
        mySubmitterPoller.refreshCache()

        # Verify that the workflow is gone from the cache
        self.assertEqual(len(mySubmitterPoller.jobDataCache), 0,
                         "Error: The job cache should be empty.  Contains: %i" % len(mySubmitterPoller.jobDataCache))
        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
WorkQueue provides functionality to queue large chunks of work,
thus acting as a buffer for the next steps in job processing

WMSpec objects are fed into the queue, split into coarse grained work units
and released when a suitable resource is found to execute them.

https://twiki.cern.ch/twiki/bin/view/CMS/WMCoreJobPool
"""

from __future__ import division, print_function

from builtins import str as newstr, bytes
from future.utils import viewitems, listvalues

import os
import threading
import time
from collections import defaultdict

from WMCore import Lexicon
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.Database.CMSCouch import CouchInternalServerError, CouchNotFoundError
from WMCore.Services.CRIC.CRIC import CRIC
from WMCore.Services.DBS.DBSReader import DBSReader
from WMCore.Services.LogDB.LogDB import LogDB
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.Services.RequestDB.RequestDBReader import RequestDBReader
from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue as WorkQueueDS
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper, getWorkloadFromTask
from WMCore.WorkQueue.DataLocationMapper import WorkQueueDataLocationMapper
from WMCore.WorkQueue.DataStructs.ACDCBlock import ACDCBlock
from WMCore.WorkQueue.DataStructs.WorkQueueElement import possibleSites
from WMCore.WorkQueue.DataStructs.WorkQueueElementsSummary import getGlobalSiteStatusSummary
from WMCore.WorkQueue.Policy.End import endPolicy
from WMCore.WorkQueue.Policy.Start import startPolicy
from WMCore.WorkQueue.WorkQueueBackend import WorkQueueBackend
from WMCore.WorkQueue.WorkQueueBase import WorkQueueBase
from WMCore.WorkQueue.WorkQueueExceptions import (TERMINAL_EXCEPTIONS, WorkQueueError, WorkQueueNoMatchingElements,
                                                  WorkQueueWMSpecError)
from WMCore.WorkQueue.WorkQueueUtils import cmsSiteNames


# Convenience constructor functions

def globalQueue(logger=None, dbi=None, **kwargs):
    """Convenience method to create a WorkQueue suitable for use globally
    """
    defaults = {'PopulateFilesets': False,
                'LocalQueueFlag': False,
                'TrackLocationOrSubscription': 'location'
                }
    defaults.update(kwargs)
    return WorkQueue(logger, dbi, **defaults)


def localQueue(logger=None, dbi=None, **kwargs):
    """Convenience method to create a WorkQueue suitable for use locally
    """
    defaults = {'TrackLocationOrSubscription': 'location'}
    defaults.update(kwargs)
    return WorkQueue(logger, dbi, **defaults)


class WorkQueue(WorkQueueBase):
    """
    _WorkQueue_

    WorkQueue object - interface to WorkQueue functionality.
    """

    def __init__(self, logger=None, dbi=None, **params):

        WorkQueueBase.__init__(self, logger, dbi)
        self.parent_queue = None
        self.params = params

        # config argument (within params) shall be reference to
        # Configuration instance
        self.config = params.get("Config", None)
        self.params.setdefault('CouchUrl', os.environ.get('COUCHURL'))
        if not self.params.get('CouchUrl'):
            raise RuntimeError('CouchUrl config value mandatory')
        self.params.setdefault('DbName', 'workqueue')
        self.params.setdefault('InboxDbName', self.params['DbName'] + '_inbox')
        self.params.setdefault('ParentQueueCouchUrl', None)  # We get work from here

        self.backend = WorkQueueBackend(self.params['CouchUrl'], self.params['DbName'],
                                        self.params['InboxDbName'],
                                        self.params['ParentQueueCouchUrl'], self.params.get('QueueURL'),
                                        logger=self.logger)
        self.workqueueDS = WorkQueueDS(self.params['CouchUrl'], self.params['DbName'],
                                       self.params['InboxDbName'])
        if self.params.get('ParentQueueCouchUrl'):
            try:
                if self.params.get('ParentQueueInboxCouchDBName'):
                    self.parent_queue = WorkQueueBackend(self.params['ParentQueueCouchUrl'].rsplit('/', 1)[0],
                                                         self.params['ParentQueueCouchUrl'].rsplit('/', 1)[1],
                                                         self.params['ParentQueueInboxCouchDBName'])
                else:
                    self.parent_queue = WorkQueueBackend(self.params['ParentQueueCouchUrl'].rsplit('/', 1)[0],
                                                         self.params['ParentQueueCouchUrl'].rsplit('/', 1)[1])
            except IndexError as ex:
                # Probable cause: Someone didn't put the global WorkQueue name in
                # the ParentCouchUrl
                msg = "Parsing failure for ParentQueueCouchUrl - probably missing dbname in input\n"
                msg += "Exception: %s\n" % str(ex)
                msg += str("ParentQueueCouchUrl: %s\n" % self.params['ParentQueueCouchUrl'])
                self.logger.error(msg)
                raise WorkQueueError(msg)
            self.params['ParentQueueCouchUrl'] = self.parent_queue.queueUrl

        # save each DBSReader instance in the class object, such that
        # the same object is not shared amongst multiple threads
        self.dbses = {}

        self.params.setdefault('QueueDepth', 1)  # when less than this locally
        self.params.setdefault('WorkPerCycle', 100)
        self.params.setdefault('RowsPerSlice', 2500)
        self.params.setdefault('MaxRowsPerCycle', 50000)
        self.params.setdefault('LocationRefreshInterval', 600)
        self.params.setdefault('FullLocationRefreshInterval', 7200)
        self.params.setdefault('TrackLocationOrSubscription', 'location')
        self.params.setdefault('ReleaseIncompleteBlocks', False)
        self.params.setdefault('ReleaseRequireSubscribed', True)
        self.params.setdefault('PopulateFilesets', True)
        self.params.setdefault('LocalQueueFlag', True)
        self.params.setdefault('QueueRetryTime', 86400)
        self.params.setdefault('stuckElementAlertTime', 172800)
        self.params.setdefault('reqmgrCompleteGraceTime', 604800)
        self.params.setdefault('cancelGraceTime', 86400)

        self.params.setdefault('JobDumpConfig', None)
        self.params.setdefault('BossAirConfig', None)

        self.params['QueueURL'] = self.backend.queueUrl  # url this queue is visible on
        # backend took previous QueueURL and sanitized it
        self.params.setdefault('WMBSUrl', None)  # this will only be set on local Queue
        if self.params.get('WMBSUrl'):
            self.params['WMBSUrl'] = Lexicon.sanitizeURL(self.params['WMBSUrl'])['url']
        self.params.setdefault('Team', "")

        if self.params.get('CacheDir'):
            try:
                os.makedirs(self.params['CacheDir'])
            except OSError:
                pass
        elif self.params.get('PopulateFilesets'):
            raise RuntimeError('CacheDir mandatory for local queue')

        if self.params.get('CRIC'):
            self.cric = self.params['CRIC']
        else:
            self.cric = CRIC()

        self.params.setdefault('SplittingMapping', {})
        self.params['SplittingMapping'].setdefault('DatasetBlock',
                                                   {'name': 'Block',
                                                    'args': {}}
                                                   )
        self.params['SplittingMapping'].setdefault('MonteCarlo',
                                                   {'name': 'MonteCarlo',
                                                    'args': {}}
                                                   )
        self.params['SplittingMapping'].setdefault('Dataset',
                                                   {'name': 'Dataset',
                                                    'args': {}}
                                                   )
        self.params['SplittingMapping'].setdefault('Block',
                                                   {'name': 'Block',
                                                    'args': {}}
                                                   )
        self.params['SplittingMapping'].setdefault('ResubmitBlock',
                                                   {'name': 'ResubmitBlock',
                                                    'args': {}}
                                                   )

        self.params.setdefault('EndPolicySettings', {})

        assert (self.params['TrackLocationOrSubscription'] in ('subscription',
                                                               'location'))
        # Can only release blocks on location
        if self.params['TrackLocationOrSubscription'] == 'location':
            if self.params['SplittingMapping']['DatasetBlock']['name'] != 'Block':
                raise RuntimeError('Only blocks can be released on location')

        self.params.setdefault('rucioAccount', "wmcore_transferor")

        self.rucio = Rucio(self.params['rucioAccount'],
                           self.params['rucioUrl'], self.params['rucioAuthUrl'],
                           configDict=dict(logger=self.logger))


        self.dataLocationMapper = WorkQueueDataLocationMapper(self.logger, self.backend,
                                                              rucio=self.rucio,
                                                              cric=self.cric,
                                                              locationFrom=self.params['TrackLocationOrSubscription'],
                                                              incompleteBlocks=self.params['ReleaseIncompleteBlocks'],
                                                              requireBlocksSubscribed=not self.params[
                                                                  'ReleaseIncompleteBlocks'],
                                                              fullRefreshInterval=self.params[
                                                                  'FullLocationRefreshInterval'],
                                                              updateIntervalCoarseness=self.params[
                                                                  'LocationRefreshInterval'])

        # used for only global WQ
        if self.params.get('ReqMgrServiceURL'):
            self.reqmgrSvc = ReqMgr(self.params['ReqMgrServiceURL'])

        if self.params.get('RequestDBURL'):
            # This is need for getting post call
            # TODO: Change ReqMgr api to accept post for for retrieving the data and remove this
            self.requestDB = RequestDBReader(self.params['RequestDBURL'])

        # set the thread name before create the log db.
        # only sets that when it is not set already
        # setLogDB

        myThread = threading.currentThread()
        if myThread.getName() == "MainThread":  # this should be only GQ case other cases thread name should be set
            myThread.setName(self.__class__.__name__)

        centralurl = self.params.get("central_logdb_url")
        identifier = self.params.get("log_reporter")
        self.logdb = LogDB(centralurl, identifier, logger=self.logger)

        self.logger.debug("WorkQueue created successfully")

    def __len__(self):
        """Returns number of Available elements in queue"""
        return self.backend.queueLength()

    def setStatus(self, status, elementIDs=None, SubscriptionId=None, WorkflowName=None):
        """
        _setStatus_, throws an exception if no elements are updated

        """
        try:
            if not elementIDs:
                elementIDs = []
            iter(elementIDs)
            if isinstance(elementIDs, (newstr, bytes)):
                raise TypeError
        except TypeError:
            elementIDs = [elementIDs]

        if status == 'Canceled':  # Cancel needs special actions
            return self.cancelWork(elementIDs, SubscriptionId, WorkflowName)

        args = {}
        if SubscriptionId:
            args['SubscriptionId'] = SubscriptionId
        if WorkflowName:
            args['RequestName'] = WorkflowName

        affected = self.backend.getElements(elementIDs=elementIDs, **args)
        if not affected:
            raise WorkQueueNoMatchingElements("No matching elements")

        for x in affected:
            x['Status'] = status
        elements = self.backend.saveElements(*affected)
        if len(affected) != len(elements):
            raise RuntimeError("Some elements not updated, see log for details")

        return elements

    def setPriority(self, newpriority, *workflowNames):
        """
        Update priority for a workflow, throw exception if no elements affected
        """
        self.logger.info("Priority change request to %s for %s", newpriority, str(workflowNames))
        affected = []
        for wf in workflowNames:
            affected.extend(self.backend.getElements(returnIdOnly=True, RequestName=wf))

        self.backend.updateElements(*affected, Priority=newpriority)

        if not affected:
            raise RuntimeError("Priority not changed: No matching elements")

    def resetWork(self, ids):
        """Put work back in Available state, from here either another queue
         or wmbs can pick it up.

         If work was Acquired by a child queue, the next status update will
         cancel the work in the child.

         Note: That the same child queue is free to pick the work up again,
          there is no permanent blacklist of queues.
        """
        self.logger.info("Resetting elements %s", str(ids))
        try:
            iter(ids)
        except TypeError:
            ids = [ids]

        return self.backend.updateElements(*ids, Status='Available',
                                           ChildQueueUrl=None, WMBSUrl=None)

    def getWork(self, jobSlots, siteJobCounts, excludeWorkflows=None):
        """
        Get available work from the queue, inject into wmbs & mark as running

        jobSlots is dict format of {site: estimateJobSlot}
        of the resources to get work for.

        siteJobCounts is a dict format of {site: {prio: jobs}}
        """
        excludeWorkflows = excludeWorkflows or []
        results = []
        if not self.backend.isAvailable():
            self.logger.warning('Backend busy or down: skipping fetching of work')
            return results

        # TODO AMR: perhaps numElems limit should be removed for LQ -> WMBS acquisition
        matches, _ = self.backend.availableWork(jobSlots, siteJobCounts,
                                                excludeWorkflows=excludeWorkflows,
                                                numElems=self.params['WorkPerCycle'],
                                                rowsPerSlice=self.params['RowsPerSlice'],
                                                maxRows=self.params['MaxRowsPerCycle'])

        self.logger.info('Got %i elements matching the constraints', len(matches))
        if not matches:
            return results

        myThread = threading.currentThread()
        # cache wmspecs for lifetime of function call, likely we will have multiple elements for same spec.
        # TODO: Check to see if we can skip spec loading - need to persist some more details to element
        wmspecCache = {}
        for match in matches:
            blockName, dbsBlock = None, None
            if self.params['PopulateFilesets']:
                if match['RequestName'] not in wmspecCache:
                    wmspec = self.backend.getWMSpec(match['RequestName'])
                    wmspecCache[match['RequestName']] = wmspec
                else:
                    wmspec = wmspecCache[match['RequestName']]

                try:
                    if match['StartPolicy'] == 'Dataset':
                        # actually returns dataset name and dataset info
                        blockName, dbsBlock = self._getDBSDataset(match)
                    elif match['Inputs']:
                        blockName, dbsBlock = self._getDBSBlock(match, wmspec)
                except Exception as ex:
                    msg = "%s, %s: \n" % (wmspec.name(), list(match['Inputs']))
                    msg += "failed to retrieve data from DBS/Rucio in LQ: \n%s" % str(ex)
                    self.logger.exception(msg)
                    self.logdb.post(wmspec.name(), msg, 'error')
                    continue

                try:
                    match['Subscription'] = self._wmbsPreparation(match,
                                                                  wmspec,
                                                                  blockName,
                                                                  dbsBlock)
                    self.logdb.delete(wmspec.name(), "error", this_thread=True)
                except Exception as ex:
                    if getattr(myThread, 'transaction', None) is not None:
                        myThread.transaction.rollback()
                    msg = "Failed to create subscription for %s with block name %s" % (wmspec.name(), blockName)
                    msg += "\nError: %s" % str(ex)
                    self.logger.exception(msg)
                    self.logdb.post(wmspec.name(), msg, 'error')
                    continue

            results.append(match)

        del wmspecCache  # remove cache explicitly
        self.logger.info('Injected %s out of %s units into WMBS', len(results), len(matches))
        return results

    def _getDbs(self, dbsUrl):
        """
        If we have already construct a DBSReader object pointing to
        the DBS URL provided, return it. Otherwise, create and return
        a new instance.
        :param dbsUrl: string with the DBS url
        :return: an instance of DBSReader
        """
        if dbsUrl in self.dbses:
            return self.dbses[dbsUrl]
        return DBSReader(dbsUrl)

    def _getDBSDataset(self, match):
        """
        Given a workqueue element with Dataset start policy, find all blocks
        with valid files and resolve their location in Rucio.
        :param match: workqueue element dictionary
        :return: a tuple of the dataset name and its files and RSEs
        """
        dbsDatasetDict = {'Files': [], 'PhEDExNodeNames': []}
        dbs = self._getDbs(match['Dbs'])
        datasetName = list(match['Inputs'])[0]

        blocks = dbs.listFileBlocks(datasetName)
        for blockName in blocks:
            blockSummary = dbs.getFileBlock(blockName)
            if not blockSummary['Files']:
                self.logger.warning("Block name %s has no valid files. Skipping it.", blockName)
                continue
            blockSummary['PhEDExNodeNames'] = self.rucio.getDataLockedAndAvailable(name=blockName,
                                                                                   account=self.params['rucioAccount'])
            dbsDatasetDict['Files'].extend(blockSummary['Files'])
            dbsDatasetDict['PhEDExNodeNames'].extend(blockSummary['PhEDExNodeNames'])

        dbsDatasetDict['PhEDExNodeNames'] = list(set(dbsDatasetDict['PhEDExNodeNames']))

        return datasetName, dbsDatasetDict

    def _getDBSBlock(self, match, wmspec):
        """Get DBS info for this block"""
        blockName = list(match['Inputs'])[0]  # TODO: Allow more than one

        if match['ACDC']:
            acdcInfo = match['ACDC']
            acdc = DataCollectionService(acdcInfo["server"], acdcInfo["database"])
            splitedBlockName = ACDCBlock.splitBlockName(blockName)
            fileLists = acdc.getChunkFiles(acdcInfo['collection'],
                                           acdcInfo['fileset'],
                                           splitedBlockName['Offset'],
                                           splitedBlockName['NumOfFiles'])

            block = {}
            block["Files"] = fileLists
            return blockName, block
        else:
            dbs = self._getDbs(match['Dbs'])
            if wmspec.getTask(match['TaskName']).parentProcessingFlag():
                dbsBlockDict = dbs.getFileBlockWithParents(blockName)
                dbsBlockDict['PhEDExNodeNames'] = self.rucio.getDataLockedAndAvailable(name=blockName,
                                                                                       account=self.params['rucioAccount'])
            elif wmspec.getRequestType() == 'StoreResults':
                dbsBlockDict = dbs.getFileBlock(blockName)
                dbsBlockDict['PhEDExNodeNames'] = dbs.listFileBlockLocation(blockName)
            else:
                dbsBlockDict = dbs.getFileBlock(blockName)
                dbsBlockDict['PhEDExNodeNames'] = self.rucio.getDataLockedAndAvailable(name=blockName,
                                                                                       account=self.params['rucioAccount'])

        return blockName, dbsBlockDict

    def _wmbsPreparation(self, match, wmspec, blockName, dbsBlock):
        """Inject data into wmbs and create subscription. """
        from WMCore.WorkQueue.WMBSHelper import WMBSHelper
        # the parent element (from local couch) can be fetch via:
        # curl -ks -X GET 'http://localhost:5984/workqueue/<ParentQueueId>'

        # Keep in mind that WQE contains sites, wmbs location contains pnns
        commonSites = possibleSites(match)
        commonLocation = self.cric.PSNstoPNNs(commonSites, allowPNNLess=True)
        msg = "Running WMBS preparation for %s with ParentQueueId %s,\n  with common location %s"
        self.logger.info(msg, match['RequestName'], match['ParentQueueId'], commonLocation)

        mask = match['Mask']
        wmbsHelper = WMBSHelper(wmspec, match['TaskName'], blockName, mask,
                                self.params['CacheDir'], commonLocation)

        sub, match['NumOfFilesAdded'] = wmbsHelper.createSubscriptionAndAddFiles(block=dbsBlock)
        self.logger.info("Created top level subscription %s for %s with %s files",
                         sub['id'], match['RequestName'], match['NumOfFilesAdded'])

        # update couch with wmbs subscription info
        match['SubscriptionId'] = sub['id']
        match['Status'] = 'Running'
        # do update rather than save to avoid conflicts from other thread writes
        self.backend.updateElements(match.id, Status='Running', SubscriptionId=sub['id'],
                                    NumOfFilesAdded=match['NumOfFilesAdded'])
        self.logger.info("LQE %s set to 'Running' for request %s", match.id, match['RequestName'])

        return sub

    def _assignToChildQueue(self, queue, *elements):
        """Assign work from parent to queue"""
        workByRequest = {}
        for ele in elements:
            ele['Status'] = 'Negotiating'
            ele['ChildQueueUrl'] = queue
            ele['ParentQueueUrl'] = self.params['ParentQueueCouchUrl']
            ele['WMBSUrl'] = self.params["WMBSUrl"]
            workByRequest.setdefault(ele['RequestName'], 0)
            workByRequest[ele['RequestName']] += 1
        work = self.parent_queue.saveElements(*elements)
        self.logger.info("Assigned work to the child queue for:")
        for reqName, numElem in viewitems(workByRequest):
            self.logger.info("    %d elements for: %s", numElem, reqName)
        return work

    def doneWork(self, elementIDs=None, SubscriptionId=None, WorkflowName=None):
        """Mark work as done
        """
        return self.setStatus('Done', elementIDs=elementIDs,
                              SubscriptionId=SubscriptionId,
                              WorkflowName=WorkflowName)

    def killWMBSWorkflows(self, reqNames):
        """
        Kill/cancel workflows in WMBS and CouchDB.
        Also update job state transition in three data sources: local couch,
        local WMBS and dashboard.
        :param reqNames: list of request names
        :return: a list of workflows that failed to be cancelled
        """
        failedWfs = []
        if not reqNames:
            return failedWfs

        # import inside function since GQ doesn't need this.
        from WMCore.WorkQueue.WMBSHelper import killWorkflow
        myThread = threading.currentThread()
        myThread.dbi = self.conn.dbi
        myThread.logger = self.logger

        for workflow in reqNames:
            try:
                self.logger.info("Killing workflow in WMBS: %s", workflow)
                killWorkflow(workflow, self.params["JobDumpConfig"], self.params["BossAirConfig"])
            except Exception as ex:
                failedWfs.append(workflow)
                msg = "Failed to kill workflow '%s' in WMBS. Error: %s" % (workflow, str(ex))
                msg += "\nIt will be retried in the next loop"
                self.logger.error(msg)
        return failedWfs

    def cancelWork(self, elementIDs=None, SubscriptionId=None, WorkflowName=None, elements=None):
        """Cancel work - delete in wmbs, delete from workqueue db, set canceled in inbox
           Elements may be directly provided or determined from series of filter arguments
        """
        if not elements:
            args = {}
            if SubscriptionId:
                args['SubscriptionId'] = SubscriptionId
            if WorkflowName:
                args['RequestName'] = WorkflowName
            elements = self.backend.getElements(elementIDs=elementIDs, **args)

        # take wf from args in case no elements exist for workflow (i.e. work was negotiating)
        requestNames = set([x['RequestName'] for x in elements]) | set([wf for wf in [WorkflowName] if wf])
        if not requestNames:
            return []
        inbox_elements = []
        for wf in requestNames:
            inbox_elements.extend(self.backend.getInboxElements(WorkflowName=wf))

        # if local queue, kill jobs, update parent to Canceled and delete elements
        if self.params['LocalQueueFlag']:
            # if we can talk to wmbs kill the jobs
            badWfsCancel = []
            if self.params['PopulateFilesets']:
                self.logger.info("Canceling work for workflow(s): %s", requestNames)
                badWfsCancel = self.killWMBSWorkflows(requestNames)
            # now we remove any wf that failed to be cancelled (and its inbox elements)
            requestNames -= set(badWfsCancel)
            for wf in badWfsCancel:
                elementsToRemove = self.backend.getInboxElements(WorkflowName=wf)
                inbox_elements = list(set(inbox_elements) - set(elementsToRemove))
            self.logger.info("New list of cancelled requests: %s", requestNames)

            # Don't update as fails sometimes due to conflicts (#3856)
            for x in inbox_elements:
                if x['Status'] != 'Canceled':
                    x.load().__setitem__('Status', 'Canceled')

            self.backend.saveElements(*inbox_elements)

        # if global queue, update non-acquired to Canceled, update parent to CancelRequested
        else:
            # Cancel in global if work has not been passed to a child queue
            elements_to_cancel = [x for x in elements if not x['ChildQueueUrl'] and x['Status'] != 'Canceled']
            # ensure all elements receive cancel request, covers case where initial cancel request missed some elements
            # without this elements may avoid the cancel and not be cleared up till they finish
            elements_not_requested = [x for x in elements if
                                      x['ChildQueueUrl'] and (x['Status'] != 'CancelRequested' and not x.inEndState())]

            self.logger.info("Canceling work for workflow(s): %s", requestNames)
            if elements_to_cancel:
                self.backend.updateElements(*[x.id for x in elements_to_cancel], Status='Canceled')
                self.logger.info("Cancel-ed element(s) %s", str([x.id for x in elements_to_cancel]))

            if elements_not_requested:
                # Don't update as fails sometimes due to conflicts (#3856)
                for x in elements_not_requested:
                    x.load().__setitem__('Status', 'CancelRequested')
                self.backend.saveElements(*elements_not_requested)
                self.logger.info("CancelRequest-ed element(s) %s", str([x.id for x in elements_not_requested]))

            inboxElemIds = [x.id for x in inbox_elements if x['Status'] != 'CancelRequested' and not x.inEndState()]
            self.backend.updateInboxElements(*inboxElemIds, Status='CancelRequested')
            # if we haven't had any updates for a while assume agent is dead and move to canceled
            if self.params.get('cancelGraceTime', -1) > 0 and elements:
                last_update = max([float(x.updatetime) for x in elements])
                if (time.time() - last_update) > self.params['cancelGraceTime']:
                    self.logger.info("%s cancelation has stalled, mark as finished", elements[0]['RequestName'])
                    # Don't update as fails sometimes due to conflicts (#3856)
                    for x in elements:
                        if not x.inEndState():
                            x.load().__setitem__('Status', 'Canceled')
                    self.backend.saveElements(*[x for x in elements if not x.inEndState()])

        return [x.id for x in elements]

    def deleteWorkflows(self, *requests):
        """Delete requests if finished"""
        for request in requests:
            request = self.backend.getInboxElements(elementIDs=[request])
            if len(request) != 1:
                raise RuntimeError('Invalid number of requests for %s' % request[0]['RequestName'])
            request = request[0]

            if request.inEndState():
                self.logger.info('Deleting request "%s" as it is %s', request.id, request['Status'])
                self.backend.deleteElements(request)
            else:
                self.logger.debug('Not deleting "%s" as it is %s', request.id, request['Status'])

    # NOTE: this function is not executed by local workqueue
    def queueWork(self, wmspecUrl, request=None, team=None):
        """
        Take and queue work from a WMSpec.

        If request name is provided but doesn't match WMSpec name
        an error is raised.

        If team is provided work will only be available to queue's
        belonging to that team.

        Duplicate specs will be ignored.
        """
        self.logger.info('queueWork() begin queueing "%s"', wmspecUrl)
        wmspec = WMWorkloadHelper()
        wmspec.load(wmspecUrl)

        if request:  # validate request name
            if request != wmspec.name():
                raise WorkQueueWMSpecError(wmspec,
                                           'Request & workflow name mismatch %s vs %s' % (request, wmspec.name()))

        # Either pull the existing inbox element or create a new one.
        try:
            inbound = self.backend.getInboxElements(elementIDs=[wmspec.name()], loadSpec=True)
            self.logger.info('Resume splitting of "%s"', wmspec.name())
        except CouchNotFoundError:
            inbound = [self.backend.createWork(wmspec, Status='Negotiating',
                                               TeamName=team, WMBSUrl=self.params["WMBSUrl"])]
            self.backend.insertElements(inbound)

        work = self.processInboundWork(inbound, throw=True)
        return len(work)

    def addWork(self, inboundElem, rucioObj=None):
        """
        Check and add new elements to an existing running request,
        if supported by the start policy.

        :param inboundElem: dict representation for a WorkQueueElement object,
            including the WMSpec file.
        :param rucioObj: object to the Rucio class
        :return: amount of new work units added to the request
        """
        result = []
        self.logger.info('Trying to add more work for: %s', inboundElem['RequestName'])

        try:
            # Check we haven't already split the work, unless it's continuous processing
            work, rejectedWork, badWork = self._splitWork(inboundElem['WMSpec'], data=inboundElem['Inputs'],
                                                          mask=inboundElem['Mask'], inbound=inboundElem,
                                                          continuous=True, rucioObj=rucioObj)

            # if there is new work, then insert it into the database
            newWork = self.backend.insertElements(work, parent=inboundElem)

            # store the inputs in the global queue inbox workflow element
            processedInputs = []
            for unit in newWork:
                processedInputs.extend(list(unit['Inputs']))

            # update the list of processed and rejected inputs with what is already
            # defined in the workqueue inbox
            processedInputs.extend(inboundElem['ProcessedInputs'])
            rejectedWork.extend(inboundElem['RejectedInputs'])
            if newWork:
                # then also update the timestamp for when new data was found
                self.backend.updateInboxElements(inboundElem.id,
                                                 ProcessedInputs=processedInputs,
                                                 RejectedInputs=rejectedWork,
                                                 TimestampFoundNewData=int(time.time()))
            # if global queue, then update workflow stats to request mgr couch doc
            # remove the "UnittestFlag" - need to create the reqmgrSvc emulator
            if not self.params.get("UnittestFlag", False):
                # get statistics for the new work. It's already validated on the server side
                totalStats = self._getTotalStats(newWork)
                self.reqmgrSvc.updateRequestStats(inboundElem['WMSpec'].name(), totalStats)

            if badWork:
                msg = "Request with the following unprocessable input data: %s" % badWork
                self.logdb.post(inboundElem['RequestName'], msg, 'warning')
        except Exception as exc:
            self.logger.error('Generic exception adding work to WQE inbox: %s. Error: %s',
                              inboundElem, str(exc))
        else:
            result.extend(newWork)

        self.logger.info('Added %d new elements for request: %s', len(result), inboundElem['RequestName'])
        return len(result)

    def status(self, status=None, elementIDs=None,
               dictKey=None, wmbsInfo=None, loadSpec=False,
               **filters):
        """
        Return elements in the queue.

        status, elementIDs & filters are 'AND'ed together to filter elements.
        dictKey returns the output as a dict with the dictKey as the key.
        wmbsInfo causes elements to be synced with their status in WMBS.
        loadSpec causes the workflow for each spec to be loaded.
        """
        items = self.backend.getElements(status=status,
                                         elementIDs=elementIDs,
                                         loadSpec=loadSpec,
                                         **filters)

        if wmbsInfo:
            self.logger.debug("Syncing element statuses with WMBS for workflow: %s", filters.get("RequestName"))
            for item in items:
                for wmbs in wmbsInfo:
                    if item['SubscriptionId'] == wmbs['subscription_id']:
                        item.updateFromSubscription(wmbs)
                        break

        # if dictKey, format as a dict with the appropriate key
        if dictKey:
            tmp = defaultdict(list)
            for item in items:
                tmp[item[dictKey]].append(item)
            items = dict(tmp)
        return items

    def getWMBSSubscriptionStatus(self):
        """
        Fetches all the subscriptions in this agent and make a summary of
        every single one of them, to be used to update WQEs
        :return: a list of dictionaries
        """
        from WMCore.WorkQueue.WMBSHelper import wmbsSubscriptionStatus
        self.logger.info("Fetching WMBS subscription status information")
        wmbsStatus = wmbsSubscriptionStatus(logger=self.logger,
                                            dbi=self.conn.dbi,
                                            conn=self.conn.getDBConn(),
                                            transaction=self.conn.existingTransaction())
        return wmbsStatus

    def statusInbox(self, status=None, elementIDs=None, dictKey=None, **filters):
        """
        Return elements in the inbox.

        status, elementIDs & filters are 'AND'ed together to filter elements.
        dictKey returns the output as a dict with the dictKey as the key.
        """
        items = self.backend.getInboxElements(status, elementIDs, **filters)

        # if dictKey, given format as a dict with the appropriate key
        if dictKey:
            tmp = defaultdict(list)
            for item in items:
                tmp[item[dictKey]].append(item)
            items = dict(tmp)

        return items

    def updateLocationInfo(self):
        """
        Update locations info for elements.
        """
        self.logger.info('Executing data location update...')
        if not self.backend.isAvailable():
            self.logger.warning('Backend busy or down: skipping location update')
            return 0
        result = self.dataLocationMapper()
        self.backend.recordTaskActivity('location_refresh')
        return result

    def _printLog(self, msg, printFlag, logLevel):
        if printFlag:
            print(msg)
        else:
            getattr(self.logger, logLevel)(msg)

    def pullWorkConditionCheck(self, printFlag=False):

        if not self.params['ParentQueueCouchUrl']:
            msg = 'Unable to pull work from parent, ParentQueueCouchUrl not provided'
            self._printLog(msg, printFlag, "warning")
            return False
        if not self.backend.isAvailable() or not self.parent_queue.isAvailable():
            msg = 'Backend busy or down: skipping work pull'
            self._printLog(msg, printFlag, "warning")
            return False

        left_over = self.parent_queue.getElements('Negotiating', returnIdOnly=True,
                                                  ChildQueueUrl=self.params['QueueURL'])
        if left_over:
            msg = 'Not pulling more work. Still replicating %d previous units, ids:\n%s' % (len(left_over), left_over)
            self._printLog(msg, printFlag, "warning")
            return False

        still_processing = self.backend.getInboxElements('Negotiating', returnIdOnly=True)
        if still_processing:
            msg = 'Not pulling more work. Still processing %d previous units' % len(still_processing)
            self._printLog(msg, printFlag, "warning")
            return False

        return True

    def freeResouceCheck(self):
        """
        This method looks into the WMBS and BossAir tables and collect
        two types of information:
         1) sites and the total slots available for job creation
         2) sites and the number of pending jobs grouped by priority
        With that information in hands, it looks at the local workqueue elements
        sitting in Available status and update the 2nd data structure (thus it
        updates number of jobs pending by priority according to the LQEs), which
        is then used to know which work can be acquired from the parent queue or not.
        :return: a tuple of dictionaries (or empty lists)
        """
        from WMCore.WorkQueue.WMBSHelper import freeSlots
        resources, jobCounts = freeSlots(self.params['QueueDepth'], knownCmsSites=cmsSiteNames())
        # now update jobCounts with work that is already available in the local queue
        _, jobCounts = self.backend.calculateAvailableWork(resources, jobCounts)

        return (resources, jobCounts)

    def getAvailableWorkfromParent(self, resources, jobCounts, printFlag=False):
        self.logger.info("Going to fetch work from the parent queue: %s", self.parent_queue.queueUrl)
        work, _ = self.parent_queue.availableWork(resources, jobCounts, self.params['Team'],
                                                  numElems=self.params['WorkPerCycle'],
                                                  rowsPerSlice=self.params['RowsPerSlice'],
                                                  maxRows=self.params['MaxRowsPerCycle'])
        if not work:
            self._printLog('No available work in parent queue.', printFlag, "warning")
        return work

    def pullWork(self, resources=None):
        """
        Pull work from another WorkQueue to be processed:
        :param resources: optional dictionary with sites and the amount
        of slots free
        """
        jobCounts = {}
        if self.pullWorkConditionCheck() is False:
            return 0

        # NOTE: resources parameter is only used by unit tests, which do
        # not use WMBS and BossAir tables
        if not resources:
            (resources, jobCounts) = self.freeResouceCheck()
        if not resources and not jobCounts:
            return 0

        work = self.getAvailableWorkfromParent(resources, jobCounts)
        if not work:
            return 0

        work = self._assignToChildQueue(self.params['QueueURL'], *work)

        return len(work)

    def closeWork(self):
        """
        Global queue service that looks for the inbox elements that are still active
        and checks whether they should be closed for new data or not.
        An element is closed automatically when one of the following conditions holds true:
        - The StartPolicy doesn't define a OpenRunningTimeout or this delay is set to 0
        - A period longer than OpenRunningTimeout has passed since the last child element
           was created or an open block was found and the StartPolicy newDataAvailable
           function returns False.

        :return: list of workqueue_inbox elements that have been closed
        """
        workflowsToClose = []
        if self.params['LocalQueueFlag']:
            # this is a Global WorkQueue only functionality
            return workflowsToClose
        if not self.backend.isAvailable():
            self.logger.warning('Backend busy or down: Can not close work at this time')
            return workflowsToClose

        workflowsToCheck = self.backend.getInboxElements(OpenForNewData=True)
        self.logger.info("Retrieved a list of %d open workflows", len(workflowsToCheck))
        currentTime = time.time()
        for element in workflowsToCheck:
            # fetch attributes from the inbox workqueue element
            startPol = element.get('StartPolicy', {})
            openRunningTimeout = startPol.get('OpenRunningTimeout', 0)
            foundNewDataTime = element.get('TimestampFoundNewData', 0)
            if not openRunningTimeout:
                self.logger.info("Workflow %s has no OpenRunningTimeout. Queuing to be closed.",
                                 element['RequestName'])
                workflowsToClose.append(element.id)
            elif (currentTime - foundNewDataTime) > openRunningTimeout:
                # then it's been too long since the last element has been found
                self.logger.info("Workflow %s has expired OpenRunningTimeout. Queuing to be closed.",
                                 element['RequestName'])
                workflowsToClose.append(element.id)

        if workflowsToClose:
            try:
                self.logger.info('Closing workflows in workqueue_inbox for: %s', workflowsToClose)
                self.backend.updateInboxElements(*workflowsToClose, OpenForNewData=False)
                msg = 'Closed inbox elements for: %s.\n' % ', '.join(workflowsToClose)
            except CouchInternalServerError as ex:
                msg = 'Failed to close workflows with a CouchInternalServerError exception. '
                msg += 'Details: {}'.format(str(ex))
                self.logger.error(msg)
            except Exception as ex:
                msg = 'Failed to close workflows with a generic exception. '
                msg += 'Details: {}'.format(str(ex))
                self.logger.exception(msg)
        else:
            msg = 'No workflows to close.\n'

        self.backend.recordTaskActivity('workclosing', msg)

        return workflowsToClose

    def deleteCompletedWFElements(self):
        """
        deletes Workflow when workflow is in finished status
        """
        deletableStates = ["completed", "closed-out", "failed",
                           "announced", "aborted-completed", "rejected",
                           "normal-archived", "aborted-archived", "rejected-archived"]

        # fetch workflows known to workqueue + workqueue_inbox and with spec attachments
        reqNames = self.backend.getWorkflows(includeInbox=True, includeSpecs=True)
        self.logger.info("Retrieved %d workflows known by WorkQueue", len(reqNames))
        requestsInfo = self.requestDB.getRequestByNames(reqNames)
        deleteRequests = []
        for key, value in viewitems(requestsInfo):
            if (value["RequestStatus"] is None) or (value["RequestStatus"] in deletableStates):
                deleteRequests.append(key)
        self.logger.info("Found %d out of %d workflows in a deletable state",
                         len(deleteRequests), len(reqNames))
        return self.backend.deleteWQElementsByWorkflow(deleteRequests)

    def performSyncAndCancelAction(self, skipWMBS):
        """
        Apply end policies to determine work status & cleanup finished work
        """
        if not self.backend.isAvailable():
            self.logger.warning('Backend busy or down: skipping cleanup tasks')
            return

        if self.params['LocalQueueFlag']:
            self.backend.fixConflicts()  # before doing anything fix any conflicts

        wf_to_cancel = []  # record what we did for task_activity
        finished_elements = []

        useWMBS = not skipWMBS and self.params['LocalQueueFlag']
        if useWMBS:
            wmbsWflowSummary = self.getWMBSSubscriptionStatus()
        else:
            wmbsWflowSummary = []
        # Get queue elements grouped by their workflow with updated wmbs progress
        # Cancel if requested, update locally and remove obsolete elements
        self.logger.info('Fetching workflow information (including inbox and specs)')
        workflowsList = self.backend.getWorkflows(includeInbox=True, includeSpecs=True)
        for wf in workflowsList:
            parentQueueDeleted = True
            try:
                elements = self.status(RequestName=wf, wmbsInfo=wmbsWflowSummary)
                parents = self.backend.getInboxElements(RequestName=wf)

                self.logger.debug("Queue %s status follows:", self.backend.queueUrl)
                results = endPolicy(elements, parents, self.params['EndPolicySettings'])
                for result in results:
                    self.logger.debug("Request %s, Status %s, Full info: %s",
                                      result['RequestName'], result['Status'], result)

                    # check for cancellation requests (affects entire workflow)
                    if result['Status'] == 'CancelRequested':
                        self.logger.info('Canceling work for workflow: %s', wf)
                        canceled = self.cancelWork(WorkflowName=wf)
                        if canceled:  # global wont cancel if work in child queue
                            wf_to_cancel.append(wf)
                            break
                    elif result['Status'] == 'Negotiating':
                        self.logger.debug("Waiting for %s to finish splitting", wf)
                        continue

                    parent = result['ParentQueueElement']
                    if parent.modified:
                        self.backend.saveElements(parent)

                    if result.inEndState():
                        if elements:
                            self.logger.debug("Request %s finished (%s)",
                                              result['RequestName'], parent.statusMetrics())
                            finished_elements.extend(result['Elements'])
                        else:
                            parentQueueDeleted = False
                        continue

                    updated_elements = [x for x in result['Elements'] if x.modified]
                    for x in updated_elements:
                        self.logger.debug("Updating progress %s (%s): %s", x['RequestName'], x.id, x.statusMetrics())
                        self.backend.updateElements(x.id, **x.statusMetrics())

                if not parentQueueDeleted:
                    self.logger.info('Waiting for parent queue to delete "%s"', wf)

            except Exception as ex:
                self.logger.error('Error processing workflow "%s": %s', wf, str(ex))

        msg = 'Finished elements: %s\nCanceled workflows: %s' % (', '.join(["%s (%s)" % (x.id, x['RequestName']) \
                                                                            for x in finished_elements]),
                                                                 ', '.join(wf_to_cancel))

        self.logger.debug(msg)
        self.backend.recordTaskActivity('housekeeping', msg)

    def performQueueCleanupActions(self, skipWMBS=False):

        try:
            self.logger.info("Deleting completed workflow WQ elements ...")
            res = self.deleteCompletedWFElements()
            self.logger.info("Deleted %d elements from workqueue/inbox database", res)
        except Exception as ex:
            self.logger.exception('Error deleting WQ elements. Details: %s', str(ex))

        try:
            self.logger.info("Syncing and cancelling work ...")
            self.performSyncAndCancelAction(skipWMBS)
        except Exception as ex:
            self.logger.error('Error syncing and canceling WQ elements. Details: %s', str(ex))

    def _splitWork(self, wmspec, data=None, mask=None, inbound=None, continuous=False, rucioObj=None):
        """
        Split work from a parent into WorkQeueueElements.

        If data param supplied use that rather than getting input data from
        wmspec. Used for instance when global splits by Block (avoids having to
        modify wmspec block whitelist - thus all appear as same wf in wmbs)

        mask can be used to specify i.e. event range.

        The inbound and continuous parameters are used to split
        and already split inbox element.
        """
        # give preference to rucio object created by the CherryPy threads
        if not rucioObj:
            rucioObj = self.rucio

        totalUnits = []
        # split each top level task into constituent work elements
        # get the acdc server and db name
        for topLevelTask in wmspec.taskIterator():
            spec = getWorkloadFromTask(topLevelTask)
            policyName = spec.startPolicy()
            if not policyName:
                raise RuntimeError("WMSpec doesn't define policyName, current value: '%s'" % policyName)

            policy = startPolicy(policyName, self.params['SplittingMapping'],
                                 rucioObj=rucioObj, logger=self.logger)
            if not policy.supportsWorkAddition() and continuous:
                # Can't split further with a policy that doesn't allow it
                continue
            if continuous:
                policy.modifyPolicyForWorkAddition(inbound)
            self.logger.info('Splitting %s with policy name %s and policy params %s',
                             topLevelTask.getPathName(), policyName,
                             self.params['SplittingMapping'].get(policyName))
            units, rejectedWork, badWork = policy(spec, topLevelTask, data, mask, continuous=continuous)
            self.logger.info('Work splitting completed with %d units, %d rejectedWork and %d badWork',
                             len(units), len(rejectedWork), len(badWork))
            for unit in units:
                msg = 'Queuing element {} for {} with policy {}, '.format(unit.id, unit['Task'].getPathName(),
                                                                          unit['StartPolicy'])
                msg += 'with {} job(s) and {} lumis'.format(unit['Jobs'], unit['NumberOfLumis'])
                if unit['Inputs']:
                    msg += ' on %s' % list(unit['Inputs'])[0]
                if unit['Mask']:
                    msg += ' on events %d-%d' % (unit['Mask']['FirstEvent'], unit['Mask']['LastEvent'])
                self.logger.info(msg)
            totalUnits.extend(units)

        return (totalUnits, rejectedWork, badWork)

    def _getTotalStats(self, units):
        totalToplevelJobs = 0
        totalEvents = 0
        totalLumis = 0
        totalFiles = 0

        for unit in units:
            totalToplevelJobs += unit['Jobs']
            totalEvents += unit['NumberOfEvents']
            totalLumis += unit['NumberOfLumis']
            totalFiles += unit['NumberOfFiles']

        return {'total_jobs': totalToplevelJobs,
                'input_events': totalEvents,
                'input_lumis': totalLumis,
                'input_num_files': totalFiles}

    def processInboundWork(self, inbound_work=None, throw=False, continuous=False, rucioObj=None):
        """Retrieve work from inbox, split and store
        If request passed then only process that request
        """
        inbound_work = inbound_work or []
        msg = "Executing processInboundWork with {} inbound_work, ".format(len(inbound_work))
        msg += "throw: {} and continuous: {}".format(throw, continuous)
        self.logger.info(msg)
        if self.params['LocalQueueFlag']:
            self.logger.info("fixing conflict...")
            self.backend.fixConflicts()  # db should be consistent

        result = []
        if not inbound_work and continuous:
            # This is not supported
            return result
        if not inbound_work:
            inbound_work = self.backend.getElementsForSplitting()
            self.logger.info('Retrieved %d elements for splitting with continuous flag: %s',
                             len(inbound_work), continuous)
        for inbound in inbound_work:
            try:
                # Check we haven't already split the work, unless it's continuous processing
                work = not continuous and self.backend.getElementsForParent(inbound)
                if work:
                    self.logger.info('Request "%s" already split - Resuming', inbound['RequestName'])
                else:
                    work, rejectedWork, badWork = self._splitWork(inbound['WMSpec'], data=inbound['Inputs'],
                                                                  mask=inbound['Mask'], inbound=inbound,
                                                                  continuous=continuous, rucioObj=rucioObj)

                    # save inbound work to signal we have completed queueing
                    # if this fails, rerunning will pick up here
                    newWork = self.backend.insertElements(work, parent=inbound)
                    # get statistics for the new work
                    totalStats = self._getTotalStats(newWork)

                    if not continuous:
                        # Update to Acquired when it's the first processing of inbound work
                        self.backend.updateInboxElements(inbound.id, Status='Acquired')

                    # store the inputs in the global queue inbox workflow element
                    if not self.params.get('LocalQueueFlag'):
                        processedInputs = []
                        for unit in work:
                            processedInputs.extend(list(unit['Inputs']))
                        self.backend.updateInboxElements(inbound.id, ProcessedInputs=processedInputs,
                                                         RejectedInputs=rejectedWork)
                        # if global queue, then update workflow stats to request mgr couch doc
                        # remove the "UnittestFlag" - need to create the reqmgrSvc emulator
                        if not self.params.get("UnittestFlag", False):
                            self.reqmgrSvc.updateRequestStats(inbound['WMSpec'].name(), totalStats)

                    if badWork:
                        msg = "Request with the following unprocessable input data: %s" % badWork
                        self.logdb.post(inbound['RequestName'], msg, 'warning')
            except TERMINAL_EXCEPTIONS as ex:
                msg = 'Terminal exception splitting WQE: %s' % inbound
                self.logger.error(msg)
                self.logdb.post(inbound['RequestName'], msg, 'error')
                if not continuous:
                    # Only fail on first splitting
                    self.logger.error('Failing workflow "%s": %s', inbound['RequestName'], str(ex))
                    self.backend.updateInboxElements(inbound.id, Status='Failed')
                    if throw:
                        raise
            except Exception as ex:
                if continuous:
                    continue
                msg = 'Exception splitting wqe %s for %s: %s' % (inbound.id, inbound['RequestName'], str(ex))
                self.logger.exception(msg)
                self.logdb.post(inbound['RequestName'], msg, 'error')

                if throw:
                    raise
                continue
            else:
                result.extend(work)

        requests = ', '.join(list(set(['"%s"' % x['RequestName'] for x in result])))
        if requests:
            self.logger.info('Split work for request(s): %s', requests)

        return result

    def getWMBSInjectionStatus(self, workflowName=None, drainMode=False):
        """
        if the parent queue exist return the result from parent queue.
        other wise return the result from the current queue.
        (In general parent queue always exist when it is called from local queue
        except T1 skim case)
        returns list of [{workflowName: injection status (True or False)}]
        if the workflow is not exist return []
        """
        if self.parent_queue and not drainMode:
            return self.parent_queue.getWMBSInjectStatus(workflowName)
        return self.backend.getWMBSInjectStatus(workflowName)

    def monitorWorkQueue(self, status=None):
        """
        Uses the workqueue data-service to retrieve a few basic information
        regarding all the elements in the queue.
        """
        status = status or []
        results = {}
        start = int(time.time())
        results['workByStatus'] = self.workqueueDS.getJobsByStatus()
        results['workByStatusAndPriority'] = self.workqueueDS.getJobsByStatusAndPriority()
        results['workByAgentAndStatus'] = self.workqueueDS.getChildQueuesAndStatus()
        results['workByAgentAndPriority'] = self.workqueueDS.getChildQueuesAndPriority()

        # now the heavy procesing for the site information
        elements = self.workqueueDS.getElementsByStatus(status)
        uniSites, posSites = getGlobalSiteStatusSummary(elements, status=status)
        results['uniqueJobsPerSiteAAA'] = uniSites
        results['possibleJobsPerSiteAAA'] = posSites
        uniSites, posSites = getGlobalSiteStatusSummary(elements, status=status, dataLocality=True)
        results['uniqueJobsPerSite'] = uniSites
        results['possibleJobsPerSite'] = posSites

        end = int(time.time())
        results["total_query_time"] = end - start
        return results



"""
MySQL implementation of File.GetParents

Return a list of lfn's which are parents for a file.
"""
from WMCore.Database.DBFormatter import DBFormatter

class GetParents(DBFormatter):
    sql = """SELECT lfn FROM wmbs_file_details WHERE id IN (
              SELECT parent FROM wmbs_file_parent WHERE child = (
                SELECT id FROM wmbs_file_details WHERE lfn = :lfn
            )
        )"""

    def getBinds(self, files=None):
        binds = []
        files = self.dbi.makelist(files)
        for f in files:
            binds.append({'lfn': f})
        return binds

    def format(self, result):
        out = []
        for r in result:
            for f in r.fetchall():
                out.append(f[0])
            r.close()
        return out

    def execute(self, files=None, conn = None, transaction = False):
        binds = self.getBinds(files)
        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)
        return self.format(result)



"""
Module defines REST API methods and their handles.
Implementation of handles is in corresponding modules, not here.

"""

import cherrypy

from WMCore.Configuration import Configuration
from WMCore.REST.Server import RESTApi
from WMCore.REST.Format import JSONFormat
from WMCore.REST.Services import ProcessMatrix

from WMCore.WMStats.Service.MetaDataInfo import ServerInfo
from WMCore.WMStats.Service.RequestInfo import (RequestInfo, FinishedStatusInfo,
                                                TeamInfo, JobDetailInfo)
from WMCore.WMStats.Service.ActiveRequestJobInfo import (ActiveRequestJobInfo, GlobalLockList,
                                                         ProtectedLFNList, ProtectedLFNListOnlyFinalOutput,
                                                         FilteredActiveRequestJobInfo)


class RestApiHub(RESTApi):
    """
    Server object for REST data access API.

    """

    def __init__(self, app, config, mount):
        """
        :arg app: reference to application object; passed to all entities.
        :arg config: reference to configuration; passed to all entities.
        :arg str mount: API URL mount point; passed to all entities."""

        RESTApi.__init__(self, app, config, mount)

        cherrypy.log("WMStats entire configuration:\n%s" % Configuration.getInstance())
        cherrypy.log("WMStats REST hub configuration subset:\n%s" % config)
        # only allows json format for return value
        self.formats = [('application/json', JSONFormat())]
        self._add({"info": ServerInfo(app, self, config, mount),
                   "teams": TeamInfo(app, self, config, mount),
                   "request": RequestInfo(app, self, config, mount),
                   "isfinished": FinishedStatusInfo(app, self, config, mount),
                   "jobdetail": JobDetailInfo(app, self, config, mount),
                   "requestcache": ActiveRequestJobInfo(app, self, config, mount),
                   "filtered_requests": FilteredActiveRequestJobInfo(app, self, config, mount),
                   "protectedlfns": ProtectedLFNList(app, self, config, mount),
                   "protectedlfns_final": ProtectedLFNListOnlyFinalOutput(app, self, config, mount),
                   "globallocks": GlobalLockList(app, self, config, mount),
                   "proc_status": ProcessMatrix(app, self, config, mount)
                   })



#!/usr/bin/env python
# encoding: utf-8

"""
File       : cms.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: CMS modules
"""

from __future__ import (division, print_function)

from WMCore.Cache.GenericDataCache import MemoryCacheStruct
from WMCore.ReqMgr.DataStructs.RequestStatus import REQUEST_STATE_LIST, REQUEST_STATE_TRANSITION
from WMCore.Services.CRIC.CRIC import CRIC
from WMCore.Services.TagCollector.TagCollector import TagCollector

# initialize TagCollector instance to be used in this module
TC = TagCollector()


def next_status(status=None):
    "Return next ReqMgr status for given status"
    if status:
        if status in REQUEST_STATE_TRANSITION:
            return REQUEST_STATE_TRANSITION[status]
        else:
            return 'N/A'
    return REQUEST_STATE_LIST


def sites():
    "Return known CMS site list from CRIC"
    try:
        # Download a list of all the sites from CRIC
        cric = CRIC()
        site_list = sorted(cric.getAllPSNs())
    except Exception as exc:
        msg = "ERROR: Could not retrieve sites from CRIC, reason: %s" % str(exc)
        raise Exception(msg)
    return site_list


# create a site cache and pnn cache 2 hour duration
SITE_CACHE = MemoryCacheStruct(5200, sites)


def pnns():
    """
    Returns all PhEDEx node names, excluding Buffer endpoints
    """
    cric = CRIC()

    try:
        pnn_list = sorted(cric.getAllPhEDExNodeNames(excludeBuffer=True))
    except Exception as exc:
        msg = "ERROR: Could not retrieve PNNs from CRIC, reason: %s" % str(exc)
        raise Exception(msg)
    return pnn_list


# create a site cache and pnn cache 2 hour duration
PNN_CACHE = MemoryCacheStruct(5200, pnns)


def site_white_list():
    "site white list, default all T1"
    t1_sites = [s for s in SITE_CACHE.getData() if s.startswith('T1_')]
    return t1_sites


def site_black_list():
    "site black list, default all T3"
    t3_sites = [s for s in SITE_CACHE.getData() if s.startswith('T3_')]
    return t3_sites


def lfn_bases():
    "Return LFN Base list"
    storeResultLFNBase = [
        "/store/backfill/1",
        "/store/backfill/2",
        "/store/data",
        "/store/mc",
        "/store/generator",
        "/store/relval",
        "/store/hidata",
        "/store/himc",
        "/store/results/analysisops",
        "/store/results/b_physics",
        "/store/results/b_tagging",
        "/store/results/b2g",
        "/store/results/e_gamma_ecal",
        "/store/results/ewk",
        "/store/results/exotica",
        "/store/results/forward",
        "/store/results/heavy_ions",
        "/store/results/higgs",
        "/store/results/jets_met_hcal",
        "/store/results/muon",
        "/store/results/qcd",
        "/store/results/susy",
        "/store/results/tau_pflow",
        "/store/results/top",
        "/store/results/tracker_dpg",
        "/store/results/tracker_pog",
        "/store/results/trigger"]
    return storeResultLFNBase


def lfn_unmerged_bases():
    "Return list of LFN unmerged bases"
    out = ["/store/unmerged", "/store/data", "/store/temp"]
    return out


def web_ui_names():
    "Return dict of web UI JSON naming conventions"
    maps = {
        "TimePerEvent": "TimePerEvent (seconds)",
        "OpenRunningTimeout": "OpenRunningTimeout (seconds)",
        "SizePerEvent": "SizePerEvent (KBytes)",
        "Memory": "Memory (MBytes)",
        "BlockCloseMaxSize": "BlockCloseMaxSize (Bytes)",
        "SoftTimeout": "SoftTimeout (seconds)",
    }
    #     maps = {"InputDataset": "Input dataset",
    #             "IncludeParents": "Include parents",
    #             "PrepID": "Optional Prep ID String",
    #             "BlockBlacklist": "Block black list",
    #             "RequestPriority": "Request priority",
    #             "TimePerEvent": "Time per event (seconds)",
    #             "RunWhitelist": "Run white list",
    #             "BlockWhitelist": "Block white list",
    #             "OpenRunningTimeout": "Open Running Timeout",
    #             "DQLUploadURL": "DQM URL",
    #             "DqmSequences": "DQM Sequences",
    #             "SizePerEvent": "Size per event (KBytes)",
    #             "ScramArch": "Architecture",
    #             "EnableHarvesting": "Enable DQM Harvesting",
    #             "DQMConfigCacheID": "DQM Config CacheID",
    #             "Memory": "Memory (MBytes)",
    #             "RunBlacklist": "Run black list",
    #             "RequestString": "Optional request ID string",
    #             "CMSSWVersion": "Software releases",
    #             "DQMUploadURL":"DQM URL",
    #             "DQMSequences": "DQM sequences",
    #             "DataPileup": "Data pileup",
    #             "FilterEfficiency": "Filter efficiency",
    #             "GlobalTag": "Global tag",
    #             "MCPileup": "MonteCarlo pileup",
    #             "PrimaryDataset": "Parimary dataset",
    #             "Acquisitionera": "Acquisition era",
    #             "CmsPath": "CMS path",
    #             "DBS": "DBS urls",
    #             "ProcessingVersion": "Processing version",
    #             "ProcessingString": "Processing string",
    #             "RequestType": "Request type",
    #             "ACDCDatabase": "ACDC database",
    #             "ACDCServer": "ACDC server",
    #             "CollectionName": "Collection name",
    #             "IgnoredOutputModules": "Ignored output modules",
    #             "InitialTaskPath": "Initial task path",
    #             "KeepStepOneOutput": "Keep step one output",
    #             "KeepStepTwoOutput": "Keep step two output",
    #             "StepOneConfigCacheID": "Step one config cache id",
    #             "StepOneOutputModuleName": "Step one output module name",
    #             "StepThreeConfigCacheID": "Step three config cache id",
    #             "StepTwoConfigCacheID": "Step two config cache id",
    #             "StepTwoOutputModuleName": "Step two output module name",
    #             "Sitewhitelist": "Site white list",
    #             "Siteblacklist": "Site black list",
    #             "Mergedlfnbase": "Merged LFN base",
    #             "Maxrss": "Max RSS",
    #             "Trustsitelists": "Trust site lists",
    #             "Unmergedlfnbase": "Unmerged LFN base",
    #             "Minmergesize": "Min merge size",
    #             "Maxmergesize": "Max merge size",
    #             "Maxmergeevents": "Max merge events",
    #             "Blockclosemaxwaittime": "Block close max wait time",
    #             "Blockclosemaxfiles": "Block close max files",
    #             "Blockclosemaxevents": "Block close max events",
    #             "Blockclosemaxsize": "Block close max size",
    #             "Softtimeout": "Soft time out",
    #             "Graceperiod": "Grace period"
    #             }
    return maps


def dqm_urls():
    "Return list of DQM urls"
    urls = [
        "https://cmsweb.cern.ch/dqm/dev",
        "https://cmsweb.cern.ch/dqm/offline",
        "https://cmsweb.cern.ch/dqm/relval",
        "https://cmsweb-testbed.cern.ch/dqm/dev",
        "https://cmsweb-testbed.cern.ch/dqm/offline",
        "https://cmsweb-testbed.cern.ch/dqm/relval",
    ]
    return urls


def dbs_urls():
    "Return list of DBS urls"
    urls = []
    base = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
    for inst in ["prod", "phys01", "phys02", "phys03"]:
        urls.append(base.replace("prod", inst))
    return urls


def couch_url():
    "Return central couch url"
    url = "https://cmsweb.cern.ch/couchdb"
    return url


def releases(arch=None):
    "Return list of CMSSW releases"
    return TC.releases(arch)


def architectures():
    "Return list of CMSSW architectures"
    return TC.architectures()


def scenarios():
    "Return list of scenarios"
    slist = ["pp", "cosmics", "hcalnzs", "preprodmc", "prodmc"]
    return slist


def cms_groups():
    "Return list of CMS data-ops groups"
    groups = ["DATAOPS"]
    return groups


def dashboardActivities():
    "Return list of dashboard activities"
    activity = ['reprocessing', 'production', 'relval', 'harvesting',
                'storeresults', 'integration', 'test']
    return activity



"""
WMBS version of the EventAwareLumiBased splitting
algorithm unittest

Created on Oct 2, 2012

@author: dballest
"""

import threading
import unittest

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.TestInit import TestInit


class EventAwareLumiBasedTest(unittest.TestCase):
    """
    _EventAwareLumiBasedTest_

    Test event based job splitting.
    """

    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)

        myThread = threading.currentThread()
        daofactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        locationAction = daofactory(classname="Locations.New")
        locationAction.execute(siteName="T1_US_FNAL", pnn="T1_US_FNAL_Disk")
        locationAction.execute(siteName="T2_CH_CERN", pnn="T2_CH_CERN")

        self.testWorkflow = Workflow(spec="spec.xml", owner="Steve",
                                     name="wf001", task="Test")
        self.testWorkflow.create()

        self.performanceParams = {'timePerEvent': 12,
                                  'memoryRequirement': 2300,
                                  'sizePerEvent': 400}

        return

    def tearDown(self):
        """
        _tearDown_

        Clear out WMBS.
        """
        self.testInit.clearDatabase()
        return

    def createSubscription(self, nFiles, lumisPerFile, twoSites=False, nEventsPerFile=100):
        """
        _createSubscription_

        Create a subscription for testing
        """

        baseName = makeUUID()

        testFileset = Fileset(name=baseName)
        testFileset.create()
        for i in range(nFiles):
            newFile = self.createFile('%s_%i' % (baseName, i), nEventsPerFile,
                                      i, lumisPerFile, 'T1_US_FNAL_Disk')
            newFile.create()
            testFileset.addFile(newFile)
        if twoSites:
            for i in range(nFiles):
                newFile = self.createFile('%s_%i_2' % (baseName, i), nEventsPerFile,
                                          i, lumisPerFile, 'T2_CH_CERN')
                newFile.create()
                testFileset.addFile(newFile)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()

        return testSubscription

    def createFile(self, lfn, events, run, lumis, location, lumiMultiplier=None):
        """
        _createFile_

        Create a file for testing
        """
        if lumiMultiplier is None:
            lumiMultiplier = run

        newFile = File(lfn=lfn, size=1000, events=events)
        lumiList = []
        for lumi in range(lumis):
            lumiList.append((lumiMultiplier * lumis) + lumi)
        newFile.addRun(Run(run, *lumiList))
        newFile.setLocation(location)
        return newFile

    def testA_FileSplitNoHardLimit(self):
        """
        _testA_FileSplitNoHardLimit_

        Simplest use case, there is only a self limit of events per job which
        the algorithm must adapt to on a file by file basis. At most
        one file per job so we don't have to pass information between files.
        """
        splitter = SplitterFactory()

        # Create 5 files with 7 lumi per file and 100 events per lumi on average.
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=7, twoSites=False,
                                                   nEventsPerFile=700)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)

        # First test, the optimal settings are 360 events per job
        # As we have files with 100 events per lumi, this will configure the splitting to
        # 3.6 lumis per job, which rounds to 3, the algorithm always approximates to the lower integer.
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=360,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 15, "There should be 15 jobs")
        for idx, job in enumerate(jobs, start=1):
            # Jobs may have 1 lumi or 2 check performance figures accordingly
            self.assertEqual(job['estimatedMemoryUsage'], 2300)
            if idx % 3 == 0:
                self.assertEqual(job['estimatedDiskUsage'], 100 * 400)
                self.assertEqual(job['estimatedJobTime'], 100 * 12)
            else:
                self.assertEqual(job['estimatedDiskUsage'], 3 * 100 * 400)
                self.assertEqual(job['estimatedJobTime'], 3 * 100 * 12)

        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=7, twoSites=False,
                                                   nEventsPerFile=700)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # Now set the average to 200 events per job
        # This results in the algorithm reducing the lumis per job to 2
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=200,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 20, "There should be 20 jobs")
        for idx, job in enumerate(jobs, start=1):
            # Jobs may have 1 lumi or 2 check performance figures accordingly
            self.assertEqual(job['estimatedMemoryUsage'], 2300)
            if idx % 4 == 0:
                self.assertEqual(job['estimatedDiskUsage'], 100 * 400)
                self.assertEqual(job['estimatedJobTime'], 100 * 12)
            else:
                self.assertEqual(job['estimatedDiskUsage'], 2 * 100 * 400)
                self.assertEqual(job['estimatedJobTime'], 2 * 100 * 12)

        # Check extremes, process a zero event files with lumis. It must be processed in one job
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=100, twoSites=False,
                                                   nEventsPerFile=0)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=5000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 0, "There are not enough events, so it should be 0 instead")

        # we close this fileset to get it moving
        fileset = testSubscription.getFileset()
        fileset.markOpen(False)

        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=5000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 5, "There should be 5 jobs")
        for job in jobs:
            self.assertEqual(job['estimatedMemoryUsage'], 2300)
            self.assertEqual(job['estimatedDiskUsage'], 0)
            self.assertEqual(job['estimatedJobTime'], 0)

        # Process files with 10k events per lumi, fallback to one lumi per job. We can't do better
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False,
                                                   nEventsPerFile=50000)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=5000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 25, "There should be 5 jobs")
        for job in jobs:
            self.assertEqual(job['estimatedMemoryUsage'], 2300)
            self.assertEqual(job['estimatedDiskUsage'], 10000 * 400)
            self.assertEqual(job['estimatedJobTime'], 10000 * 12)

    def testB_NoFileSplitNoHardLimit(self):
        """
        _testB_NoFileSplitNoHardLimit_

        In this case we don't split on file boundaries, check different combination of files
        make sure we make the most of the splitting, e.g. include many zero event files in
        a single job.
        """
        splitter = SplitterFactory()

        # Create 100 files with 7 lumi per file and 0 events per lumi on average.
        testSubscription = self.createSubscription(nFiles=100, lumisPerFile=7, twoSites=False,
                                                   nEventsPerFile=0)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)

        # First test, the optimal settings are 360 events per job
        # As we have files with 0 events per lumi, this will configure the splitting to
        # a single job containing all files
        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=360,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 0, "There aren't enough events, so it should have 0 job groups")

        # we close this fileset to get it moving
        fileset = testSubscription.getFileset()
        fileset.markOpen(False)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=360,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 1, "There should be 1 job")
        self.assertEqual(len(jobs[0]['input_files']), 100, "All 100 files must be in the job")
        self.assertEqual(jobs[0]['estimatedMemoryUsage'], 2300)
        self.assertEqual(jobs[0]['estimatedDiskUsage'], 0)
        self.assertEqual(jobs[0]['estimatedJobTime'], 0)

        # Create 7 files, each one with different lumi/event distributions
        testFileset = Fileset(name="FilesetA")
        testFileset.create()
        testFileA = self.createFile("/this/is/file1", 250, 0, 5, "T2_CH_CERN")
        testFileB = self.createFile("/this/is/file2", 600, 1, 1, "T2_CH_CERN")
        testFileC = self.createFile("/this/is/file3", 1200, 2, 2, "T2_CH_CERN")
        testFileD = self.createFile("/this/is/file4", 100, 3, 1, "T2_CH_CERN")
        testFileE = self.createFile("/this/is/file5", 30, 4, 1, "T2_CH_CERN")
        testFileF = self.createFile("/this/is/file6", 10, 5, 1, "T2_CH_CERN")
        testFileG = self.createFile("/this/is/file7", 151, 6, 3, "T2_CH_CERN")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.addFile(testFileD)
        testFileset.addFile(testFileE)
        testFileset.addFile(testFileF)
        testFileset.addFile(testFileG)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # Optimal settings are: jobs with 150 events per job
        # This means, the first file must be splitted in 3 lumis per job which would leave room
        # for another lumi in the second job, but the second file has a lumi too big for that
        # The 3rd job only contains the second file, the fourth and fifth job split the third file
        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=150,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 7, "7 jobs must be in the jobgroup")
        self.assertEqual(jobs[0]["mask"].getRunAndLumis(), {0: [[0, 2]]}, "Wrong mask for the first job")
        self.assertEqual(jobs[0]["estimatedJobTime"], 150 * 12)
        self.assertEqual(jobs[0]["estimatedDiskUsage"], 150 * 400)
        self.assertEqual(jobs[1]["mask"].getRunAndLumis(), {0: [[3, 4]]}, "Wrong mask for the second job")
        self.assertEqual(jobs[1]["estimatedJobTime"], 100 * 12)
        self.assertEqual(jobs[1]["estimatedDiskUsage"], 100 * 400)
        self.assertEqual(jobs[2]["mask"].getRunAndLumis(), {1: [[1, 1]]}, "Wrong mask for the third job")
        self.assertEqual(jobs[2]["estimatedJobTime"], 600 * 12)
        self.assertEqual(jobs[2]["estimatedDiskUsage"], 600 * 400)
        self.assertEqual(jobs[3]["mask"].getRunAndLumis(), {2: [[4, 4]]}, "Wrong mask for the fourth job")
        self.assertEqual(jobs[3]["estimatedJobTime"], 600 * 12)
        self.assertEqual(jobs[3]["estimatedDiskUsage"], 600 * 400)
        self.assertEqual(jobs[4]["mask"].getRunAndLumis(), {2: [[5, 5]]}, "Wrong mask for the fifth job")
        self.assertEqual(jobs[4]["estimatedJobTime"], 600 * 12)
        self.assertEqual(jobs[4]["estimatedDiskUsage"], 600 * 400)
        self.assertEqual(jobs[5]["mask"].getRunAndLumis(),
                         {3: [[3, 3]], 4: [[4, 4]], 5: [[5, 5]]}, "Wrong mask for the sixth job")
        self.assertEqual(jobs[5]["estimatedJobTime"], 140 * 12)
        self.assertEqual(jobs[5]["estimatedDiskUsage"], 140 * 400)
        self.assertEqual(jobs[6]["mask"].getRunAndLumis(), {6: [[18, 20]]}, "Wrong mask for the seventh job")
        self.assertEqual(jobs[6]["estimatedJobTime"], 150 * 12)
        self.assertEqual(jobs[6]["estimatedDiskUsage"], 150 * 400)

        for job in jobs:
            self.assertEqual(job["estimatedMemoryUsage"], 2300)
        # Test interactions of this algorithm with splitOnRun = True
        # Make 2 files, one with 3 runs and a second one with the last run of the first
        fileA = File(lfn="/this/is/file1a", size=1000,
                     events=2400)
        lumiListA = []
        lumiListB = []
        lumiListC = []
        for lumi in range(8):
            lumiListA.append(1 + lumi)
            lumiListB.append(1 + lumi)
            lumiListC.append(1 + lumi)
        fileA.addRun(Run(1, *lumiListA))
        fileA.addRun(Run(2, *lumiListA))
        fileA.addRun(Run(3, *lumiListA))
        fileA.setLocation("T1_US_FNAL_Disk")

        fileB = self.createFile('/this/is/file2a', 200, 3, 5, "T1_US_FNAL_Disk")

        testFileset = Fileset(name='FilesetB')
        testFileset.create()
        testFileset.addFile(fileA)
        testFileset.addFile(fileB)
        testFileset.commit()
        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # The settings for this splitting are 700 events per job
        jobGroups = jobFactory(splitOnRun=True,
                               halt_job_on_file_boundaries=False,
                               events_per_job=700,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")
        self.assertEqual(jobs[0]["estimatedJobTime"], 700 * 12)
        self.assertEqual(jobs[0]["estimatedDiskUsage"], 700 * 400)
        self.assertEqual(jobs[1]["estimatedJobTime"], 100 * 12)
        self.assertEqual(jobs[1]["estimatedDiskUsage"], 100 * 400)
        self.assertEqual(jobs[2]["estimatedJobTime"], 700 * 12)
        self.assertEqual(jobs[2]["estimatedDiskUsage"], 700 * 400)
        self.assertEqual(jobs[3]["estimatedJobTime"], 100 * 12)
        self.assertEqual(jobs[3]["estimatedDiskUsage"], 100 * 400)
        self.assertEqual(jobs[4]["estimatedJobTime"], 700 * 12)
        self.assertEqual(jobs[4]["estimatedDiskUsage"], 700 * 400)
        self.assertEqual(jobs[5]["estimatedJobTime"], 300 * 12)
        self.assertEqual(jobs[5]["estimatedDiskUsage"], 300 * 400)

    def testC_HardLimitSplitting(self):
        """
        _testC_HardLimitSplitting_

        Test that we can specify a event limit, the
        algorithm shall take single lumi files with more events than the limit
        and mark them for failure
        """
        splitter = SplitterFactory()

        # Create 3 files, the one in the middle is a "bad" file
        testFileset = Fileset(name="FilesetA")
        testFileset.create()
        testFileA = self.createFile("/this/is/file1", 1000, 0, 5, "T1_US_FNAL_Disk")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "T1_US_FNAL_Disk")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 2, "T1_US_FNAL_Disk")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")
        self.assertTrue(jobs[3]['failedOnCreation'], "The job processing the second file should me marked for failure")
        self.assertEqual(jobs[3]['failedReason'],
                         'File /this/is/file2 has a single lumi 1, in run 1 with too many events 1000 and it woud take 12000 sec to run')

    def testD_HardLimitSplittingOnly(self):
        """
        _testD_HardLimitSplittingOnly_

        Checks that we can split a set of files where every file has a single
        lumi too big to fit in a runnable job
        """
        splitter = SplitterFactory()

        # Create 3 single-big-lumi files
        testFileset = Fileset(name="FilesetA")
        testFileset.create()
        testFileA = self.createFile("/this/is/file1", 1000, 0, 1, "T1_US_FNAL_Disk")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "T1_US_FNAL_Disk")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 1, "T1_US_FNAL_Disk")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 3, "Three jobs must be in the jobgroup")
        for i in range(1, 4):
            self.assertTrue(jobs[i - 1]['failedOnCreation'],
                            "The job processing the second file should me marked for failure")
            error = 'File /this/is/file%s has a single lumi %d, in run %s' % (i, i - 1, i - 1)
            error += ' with too many events 1000 and it woud take 12000 sec to run'
            self.assertEqual(jobs[i - 1]['failedReason'], error)

        return
    def testE_DisableHardLimitSplitting(self):
        """
        _testC_DisableHardLimitSplitting_
        Test that we can bypass and event limit when allowCreationFailure is
        set to False. The algorithm shall take single lumi files with more events
        than the limit but not mark them for failure
        """
        splitter = SplitterFactory()

        # Create 3 files, the one in the middle is a "bad" file
        testFileset = Fileset(name="FilesetA")
        testFileset.create()
        testFileA = self.createFile("/this/is/file1", 1000, 0, 5, "T1_US_FNAL_Disk")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "T1_US_FNAL_Disk")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 2, "T1_US_FNAL_Disk")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               allowCreationFailure=False,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")
        failedJobs = [job for job in jobs if job.get('failedOnCreation', False)]
        self.assertEqual(len(failedJobs), 0, "There should be no failed jobs")

    def test_NotEnoughEvents(self):
        """
        _test_NotEnoughEvents_

        Checks whether jobs are not created when there are not enough files (actually, events)
        according to the events_per_job requested to the splitter algorithm
        """
        splitter = SplitterFactory()

        # Very small fileset (single file) without enough events
        testSubscription = self.createSubscription(nFiles=1, lumisPerFile=2, nEventsPerFile=200)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=500,
                               performance=self.performanceParams,
                               splitOnRun=False)

        self.assertEqual(len(jobGroups), 0)

        # Still a small fileset (two files) without enough events
        testSubscription = self.createSubscription(nFiles=2, lumisPerFile=2, nEventsPerFile=200)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=500,
                               performance=self.performanceParams,
                               splitOnRun=False)

        self.assertEqual(len(jobGroups), 0)

        # Finally an acceptable fileset size (three files) with enough events
        testSubscription = self.createSubscription(nFiles=3, lumisPerFile=2, nEventsPerFile=200)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=500,
                               performance=self.performanceParams,
                               splitOnRun=False)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2)
        self.assertEqual(len(jobs[0]['input_files']), 3)
        self.assertEqual(len(jobs[1]['input_files']), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 1]], 1: [[2, 3]], 2: [[4, 4]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {2: [[5, 5]]})

        # Test fileset with a single run and splitOnRun=True
        testFileset = Fileset(name="FilesetA")
        testFileA = self.createFile("/this/is/file1", 200, 1, 2, "T1_US_FNAL_Disk", lumiMultiplier=0)
        testFileB = self.createFile("/this/is/file2", 200, 1, 2, "T1_US_FNAL_Disk", lumiMultiplier=1)
        testFileC = self.createFile("/this/is/file3", 200, 1, 2, "T1_US_FNAL_Disk", lumiMultiplier=2)
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        testSubscription.create()

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=500,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2)
        self.assertEqual(len(jobs[0]['input_files']), 3)
        self.assertEqual(len(jobs[1]['input_files']), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {1: [[0, 1], [2, 3], [4, 4]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {1: [[5, 5]]})

        return


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_Dummy_

A dummy class to mimic the component for tets.
"""

import logging
import os
import threading

class DBCoreDummy:
    def __init__(self):
        self.connectUrl = os.getenv("DATABASE")

class ConfigDummy:
    def __init__(self):
        self.CoreDatabase = DBCoreDummy()

class Dummy():
    """
    _Dummy_

    A dummy class to mimic the component for tests.
    """

    def __init__(self):
        logging.debug("I am a dummy object!")
        self.config = ConfigDummy()

    def prepareToStop(self):
        """Copied from Agent.Harness"""
        # Stop all worker threads
        logging.info(">>>Terminating worker threads")
        myThread = threading.currentThread()
        myThread.workerThreadManager.terminateWorkers()



from future import standard_library
standard_library.install_aliases()

from future.utils import viewitems

import hashlib
import hmac
import urllib.parse

from http.client import HTTPConnection

from Utils.Utilities import decodeBytesToUnicodeConditional, encodeUnicodeToBytesConditional
from Utils.PythonVersion import PY3

from WMCore.WebTools.Page import make_rfc_timestamp


def makeRequest(url, values=None, verb='GET', accept="text/plain",
                contentType=None, secure=False, secureParam={}):
    """
    :rtype: (bytes (both py2 and py3), int, str (native), 
            (instance httplib.HTTPResponse in py2, 'http.client.HTTPResponse' in py3))
    """
    headers = {}
    contentType = contentType or "application/x-www-form-urlencoded"
    headers = {"content-type": contentType,
               "Accept": accept,
               "cms-auth-status": "NONE"}
    if secure:
        headers.update({"cms-auth-status": "OK",
                        "cms-authn-dn": "/DC=ch/OU=Organic Units/OU=Users/CN=Fake User",
                        "cms-authn-name": "Fake User",
                        "cms-authz-%s" % secureParam['role']:
                            "group:%s site:%s" % (secureParam['group'],
                                                  secureParam['site'])})
        headers["cms-authn-hmac"] = _generateHash(secureParam["key"], headers)

    data = None
    if verb == 'GET' and values:
        data = urllib.parse.urlencode(values, doseq=True)
    elif verb != 'GET' and values:
        # needs to test other encoding type
        if contentType == "application/x-www-form-urlencoded":
            data = urllib.parse.urlencode(values)
        else:
            # for other encoding scheme values assumed to be encoded already
            data = values
    parser = urllib.parse.urlparse(url)
    uri = parser.path
    if parser.query:
        uri += "?" + parser.query

    if verb == 'GET' and data != None:
        uri = '%s?%s' % (uri, data)

    # need to specify Content-length for POST method
    # TODO: this function needs refactoring - too verb-related branching
    if verb != 'GET':
        if data:
            headers.update({"content-length": len(data)})
        else:
            headers.update({"content-length": 0})

    conn = HTTPConnection(parser.netloc)
    conn.connect()
    conn.request(verb, uri, data, headers)
    response = conn.getresponse()
    data = response.read()
    conn.close()
    cType = response.getheader('content-type').split(';')[0]
    # data returned could be something a json like: b'"foo"', so we need to properly load it
    #if '/json' in accept:
    #    data = json.loads(data)
    return data, response.status, cType, response


def methodTest(verb, url, request_input={}, accept='text/json', contentType=None,
               output={}, expireTime=0, secure=False, secureParam={}):
    data, code, content_type, response = makeRequest(url, request_input, verb,
                                                     accept, contentType,
                                                     secure, secureParam)

    data = decodeBytesToUnicodeConditional(data, condition=PY3)

    keyMap = {'code': code, 'data': data, 'type': content_type, 'response': response}
    for key, value in viewitems(output):
        msg = 'Got a return %s != %s (got %s, type %s) (expected %s, type %s)' \
              % (keyMap[key], value, keyMap[key], type(keyMap[key]), value, type(value))
        assert keyMap[key] == value, msg

    expires = response.getheader('Expires')
    if expireTime != 0:
        timeStamp = make_rfc_timestamp(expireTime)
        assert expires == timeStamp, \
            'Expires header incorrect (%s) != (%s)' % (expires, timeStamp)

    return data, expires


def _generateHash(keyfile, headers):
    prefix = suffix = ""
    hkeys = sorted(headers.keys())
    for hk in hkeys:
        hk = hk.lower()
        if hk[0:9] in ["cms-authn", "cms-authz"]:
            prefix += "h%xv%x" % (len(hk), len(headers[hk]))
            suffix += "%s%s" % (hk, headers[hk])

    return hmac.new(keyfile, prefix + "#" + suffix, hashlib.sha1).hexdigest()



#!/usr/bin/env python
"""
_WMBSBase_

Generic methods used by all of the WMBS classes.
"""

from builtins import object

import threading

class WorkQueueBase(object):
    """
    Generic methods used by all of the WMBS classes.
    """
    def __init__(self, logger=None, dbi=None):
        """
        ___init___

        Initialize all the database connection attributes and the logging
        attritbutes.  Create a DAO factory for WMCore.WorkQueue as well. Finally,
        check to see if a transaction object has been created.  If none exists,
        create one but leave the transaction closed.
        """
        # only load dbi connection if we need it
        if dbi or 'dbi' in dir(threading.currentThread()):
            from WMCore.WMConnectionBase import WMConnectionBase
            self.conn = WMConnectionBase(daoPackage = "WMCore",
                                         logger = logger, dbi = dbi)
            self.logger = self.conn.logger
        else:
            self.conn = None
            if logger:
                self.logger = logger
            elif 'logger' in dir(threading.currentThread()):
                self.logger = threading.currentThread().logger
            else:
                import logging
                self.logger = logging.getLogger()



#!/usr/bin/env python
"""
_MarkNewFinishedSubscriptions_

MySQL implementation of SubscriptionsMarkNewFinishedSubscriptions
"""

import time

from WMCore.Database.DBFormatter import DBFormatter

class MarkNewFinishedSubscriptions(DBFormatter):
    """

    Searches for all subscriptions where the fileset is closed,
    the workflow is injected, there are no available files,
    no acquired files, no unfinished child subscriptions, no jobs that
    are not in state 'cleanout' and whose state has not changed in some time  and
    where the state is not finished. Also checks that the files in the input fileset
    are not parents of files associated with a unfinished subscription, possibly in another
    workflow.

    It marks such subscriptions and finished
    """

    completeNonJobSQL = """
                    SELECT distinct wmbs_subscription.id,
                                    wmbs_subscription.fileset,
                                    wmbs_workflow.name
                        FROM wmbs_subscription
                        INNER JOIN wmbs_fileset ON
                            wmbs_fileset.id = wmbs_subscription.fileset AND
                            wmbs_fileset.open = 0
                        INNER JOIN wmbs_workflow ON
                            wmbs_workflow.id = wmbs_subscription.workflow AND
                            wmbs_workflow.injected = 1
                        LEFT OUTER JOIN wmbs_sub_files_available ON
                            wmbs_sub_files_available.subscription = wmbs_subscription.id
                        LEFT OUTER JOIN wmbs_sub_files_acquired ON
                            wmbs_sub_files_acquired.subscription = wmbs_subscription.id
                     WHERE wmbs_subscription.finished = 0 AND
                            wmbs_sub_files_available.subscription is Null AND
                            wmbs_sub_files_acquired.subscription is Null
                    """
    subWithUnfinishedJobSQL = """
                    SELECT distinct wmbs_subscription.id
                        FROM wmbs_subscription
                        INNER JOIN wmbs_fileset ON
                            wmbs_fileset.id = wmbs_subscription.fileset AND
                            wmbs_fileset.open = 0
                        INNER JOIN wmbs_workflow ON
                            wmbs_workflow.id = wmbs_subscription.workflow AND
                            wmbs_workflow.injected = 1
                        LEFT OUTER JOIN wmbs_sub_files_available ON
                            wmbs_sub_files_available.subscription = wmbs_subscription.id
                        LEFT OUTER JOIN wmbs_sub_files_acquired ON
                            wmbs_sub_files_acquired.subscription = wmbs_subscription.id

                        INNER JOIN wmbs_jobgroup ON
                            wmbs_jobgroup.subscription = wmbs_subscription.id
                        INNER JOIN wmbs_job ON
                            wmbs_job.jobgroup = wmbs_jobgroup.id AND
                            wmbs_job.state_time > :maxTime AND
                            wmbs_job.state != %d
                     WHERE wmbs_subscription.finished = 0 AND
                        wmbs_sub_files_available.subscription is Null AND
                        wmbs_sub_files_acquired.subscription is Null
                  """

    sql = """ UPDATE wmbs_subscription
             SET wmbs_subscription.finished = 1, wmbs_subscription.last_update = :timestamp
             WHERE wmbs_subscription.id IN (
             SELECT id FROM (
                    SELECT complete_subscription.id
                        FROM ( %s ) complete_subscription
                    WHERE complete_subscription.id
                        NOT IN ( %s )
                    GROUP BY complete_subscription.id) deletable_subscriptions )""" % (
                                                        completeNonJobSQL, subWithUnfinishedJobSQL)


    def execute(self, state, timeOut = None, conn = None, transaction = False):

        currentTime = int(time.time())
        if timeOut == None:
            timeOut = currentTime

        binds = {'maxTime'   : currentTime - timeOut,
                 'timestamp' : currentTime}

        self.dbi.processData(self.sql % state,
                             binds, conn = conn,
                             transaction = transaction)

        return




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Workflow object representing how growing workflows have to be dealt
with in the input data placement (MSTransferor)
"""

from WMCore.MicroService.MSTransferor.DataStructs.Workflow import Workflow
from WMCore.Services.Rucio.RucioUtils import GROUPING_DSET


class GrowingWorkflow(Workflow):
    """
    Class to represent a Growing workflow in the context
    of input data placement in MSTransferor
    """

    def getInputData(self):
        """
        Returns the primary dataset name to be locked and
        transferred with Rucio, instead of a list of blocks.

        :return: a list with the input dataset name and an integer
                 with their total size
        """
        inputContainer = [self.getInputDataset()]
        totalBlockSize = sum([blockInfo['blockSize'] for blockInfo in self.getPrimaryBlocks().values()])

        # the whole container must be locked
        return inputContainer, totalBlockSize

    def getRucioGrouping(self):
        """
        Returns the rucio rule grouping for growing workflows.
        Input blocks can be scattered all over the provided
        RSE expression.

        :return: a string with the required DID grouping
        """
        return GROUPING_DSET



#!/usr/bin/env python
"""
_JobFactory_t_

Test the job splitting Job Factory
"""




import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow

from WMCore.JobSplitting.JobFactory import JobFactory

class JobFactoryTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testMetaData(self):
        """
        _testMetaData_

        Make sure that the workflow name, task, owner and white and black lists
        make it into each job object.
        """
        testWorkflow = Workflow(spec = "spec.pkl", owner = "Steve",
                                name = "TestWorkflow", task = "TestTask")

        testFileset = Fileset(name = "TestFileset")
        testFile = File(lfn = "someLFN")
        testFileset.addFile(testFile)
        testFileset.commit()

        testSubscription = Subscription(fileset = testFileset,
                                        workflow = testWorkflow,
                                        split_algo = "FileBased")

        myJobFactory = JobFactory(subscription = testSubscription)
        testJobGroups =  myJobFactory(siteWhitelist = ["site1"], siteBlacklist = ["site2"])
        self.assertTrue(len(testJobGroups) > 0)

        for testJobGroup in testJobGroups:
            self.assertTrue(len(testJobGroup.jobs) > 0)
            for job in testJobGroup.jobs:
                self.assertEqual(job["task"], "TestTask", "Error: Task is wrong.")
                self.assertEqual(job["workflow"], "TestWorkflow", "Error: Workflow is wrong.")
                self.assertEqual(job["owner"], "Steve", "Error: Owner is wrong.")
        return

    def testProductionRunNumber(self):
        """
        _testProductionRunNumber_

        Verify that jobs created by production subscritpions have the correct
        run number is their job mask.  Also verify that non-production
        subscriptions don't have modified run numbers.
        """
        testWorkflow = Workflow(spec = "spec.pkl", owner = "Steve",
                                name = "TestWorkflow", task = "TestTask")

        testFileset = Fileset(name = "TestFileset")
        testFile = File(lfn = "someLFN")
        testFileset.addFile(testFile)
        testFileset.commit()

        testSubscription = Subscription(fileset = testFileset,
                                        workflow = testWorkflow,
                                        split_algo = "FileBased",
                                        type = "Production")

        myJobFactory = JobFactory(subscription = testSubscription)
        testJobGroups =  myJobFactory()

        self.assertTrue(len(testJobGroups) > 0)
        for testJobGroup in testJobGroups:
            self.assertTrue(len(testJobGroup.jobs) > 0)
            for job in testJobGroup.jobs:
                self.assertEqual(job["mask"]["FirstRun"], 1, "Error: First run is wrong.")
                self.assertEqual(job["mask"]["LastRun"], 1, "Error: Last run is wrong.")

        testSubscription = Subscription(fileset = testFileset,
                                        workflow = testWorkflow,
                                        split_algo = "FileBased",
                                        type = "Processing")

        myJobFactory = JobFactory(subscription = testSubscription)
        testJobGroups =  myJobFactory()

        for testJobGroup in testJobGroups:
            for job in testJobGroup.jobs:
                self.assertEqual(job["mask"]["FirstRun"], None, "Error: First run is wrong.")
                self.assertEqual(job["mask"]["LastRun"], None, "Error: Last run is wrong.")

        return

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_Fileset_t_

Testcase for Fileset

"""
from builtins import range
import logging
import random
import unittest

from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.File import File
from WMCore.DataStructs.Run import Run

class FilesetTest (unittest.TestCase):
    """
    _FilesetTest_

    Testcase for Fileset
    """
    def setUp(self):
        """
        Create a dummy fileset and populate it with random files,
        in order to use it for the testcase methods

        """
        logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename=__file__.replace('.py','.log'),
                    filemode='w')
        self.logger = logging.getLogger('FilesetClassTest')

        #Setup the initial testcase environment:
        initialfile = File('/tmp/lfn1',1000,1,1,1)
        self.initialSet = set()
        self.initialSet.add(initialfile)

        #Create a Fileset, containing a initial file on it.
        self.fileset = Fileset(name = 'self.fileset', files = self.initialSet)
        #Populate the fileset with random files
        for i in range(1,1000):
            lfn = '/store/data/%s/%s/file.root' % (random.randint(1000, 9999),
                                              random.randint(1000, 9999))
            size = random.randint(1000, 2000)
            events = 1000
            run = random.randint(0, 2000)
            lumi = random.randint(0, 8)

            file = File(lfn=lfn, size=size, events=events, checksums = {"cksum": "1"})
            file.addRun(Run(run, *[lumi]))
            self.fileset.addFile(file)

    def tearDown(self):
        """
            No tearDown method for this testcase

        """
        pass

    def testAddFile(self):
        """
        _testAddFile_

        Testcase for the addFile method of the Fileset class

        """

        # First test - Add file and check if it's there
        testfile = File('/tmp/lfntest', 9999, 9, 9)
        self.fileset.addFile(testfile)
        self.assertTrue(testfile in self.fileset.listNewFiles(),
                        "Couldn't add file to fileset - fileset.addfile method not working")

        # Second test - Add file that was already at Fileset.files, and check if it gets updated
        testFileSame = File('/tmp/lfntest', 9999, 9, 9)
        testFileSame.setLocation(set('dummyse.dummy.com'))
        self.fileset.addFile(testFileSame)
        self.assertTrue(testFileSame in self.fileset.getFiles(),
                        'Same file copy ailed - fileset.addFile not updating location of already existing files')
        self.assertTrue(testfile in self.fileset.getFiles(),
                        'Same file copy failed - fileset.addFile unable to remove previous file from list')

        # Third test - Add file that was already at Fileset.newfiles, and check if it gets updated
        self.assertTrue(testFileSame in self.fileset.listNewFiles(),
                        'Same file copy failed - fileset.addFile not adding file to fileset.newFiles')

    def testListFiles(self):
        """
        _testListFiles_

        Testcase for the getFiles() method of the Fileset class
        """
        filesettemp = self.fileset.getFiles()
        for x in filesettemp:
            assert x in (self.fileset.files | self.fileset.newfiles), \
            'Missing file %s from file list returned from fileset.ListFiles' % x["lfn"]

    def testSetFiles(self):
        """
        Check that all files returned by the set are the same as those added to
        the fileset
        """
        filesettemp = self.fileset.getFiles(type = "set")
        for x in filesettemp:
            assert x in (self.fileset.files | self.fileset.newfiles), \
            'Missing file %s from file list returned from fileset.ListFiles' % x["lfn"]

    def testSetListCompare(self):
        """
        Test that all files in fileset.setFiles are in fileset.getFiles()
        """
        thelist = self.fileset.getFiles()
        theset = self.fileset.getFiles(type = "set")
        for x in thelist:
            assert x in (theset), \
            'Missing file %s from file list returned from fileset.ListFiles' % x["lfn"]

    def testSorting(self):
        """
        Fileset.getFiles() should be sorted the same as Fileset.getFiles(type = "lfn"),
        assert that this is the case here.
        """
        files = self.fileset.getFiles()
        lfns = self.fileset.getFiles(type = "lfn")
        for x in files:
            assert x["lfn"] in (lfns), \
            'Missing file %s from file list returned ' % x["lfn"]
            assert lfns[files.index(x)] == x["lfn"], \
            'Sorting not consistent: lfn = %s, file = %s' % (lfns[files.index(x)], x["lfn"])

    def testListLFNs(self):
        """
            Testcase for the listLFN method of the Fileset class

        """
        #Kinda slow way of verifying if the raw LFN from each file at the
        #fileset is returned from the meth
        allFiles = self.fileset.getFiles()

        #For each file returned by method listFiles, it checks if LFN is
        #present at the output of method listLFNs
        for x in allFiles:
            assert x['lfn'] in self.fileset.getFiles(type = "lfn"), 'Missing %s from ' \
            'list returned from fileset.ListLFNs' % x["lfn"]
        #Im a bit confused with this method, leave to discuss it at the meeting with Simon

    def testListNewFiles(self):
        """
            Testcase for the listNewFiles method of the Fileset class

        """
        newfilestemp = self.fileset.listNewFiles()
        assert newfilestemp == self.fileset.newfiles, 'Missing files from ' \
               'list returned from fileset.ListNewFiles'
    def testCommit(self):
        """
            Testcase for the commit method of the Fileset class

        """
        localTestFileSet = Fileset('LocalTestFileset', self.initialSet)
        fsSize = len(localTestFileSet.getFiles(type = "lfn"))
        #Dummy file to test
        fileTestCommit = File('/tmp/filetestcommit',0000,1,1)
        #File is added to the newfiles attribute of localTestFileSet
        localTestFileSet.addFile(fileTestCommit)
        assert fsSize == len(localTestFileSet.getFiles(type = "lfn")) - 1, 'file not added'\
                'correctly to test fileset'
        newfilestemp = localTestFileSet.newfiles
        assert fileTestCommit in newfilestemp, 'test file not in the new files'\
                'list'
        #After commit, dummy file is supposed to move from newfiles to files
        localTestFileSet.commit()
        #First, testing if the new file is present at file set object attribute of the Fileset object

        assert newfilestemp.issubset(localTestFileSet.files), 'Test file not ' \
                'present at fileset.files - fileset.commit ' \
                'not working properly'
        #Second, testing if the newfile set object attribute is empty
        assert localTestFileSet.newfiles == set(), \
                'Test file not present at fileset.newfiles ' \
                '- fileset.commit not working properly'

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
Group_t.py

Created by Dave Evans on 2010-07-29.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

import unittest
import os

from WMCore.GroupUser.Group import Group

class Group_t(unittest.TestCase):

    def setUp(self):
        self.database = "groupuser"
        self.url = os.getenv("COUCHURL", "http://127.0.0.1:5984")


    def testA(self):
        """instantiate and jsonise"""

        g1 = Group(name = "DMWM", administrators = ['evansde77', 'drsm79'])

        g1.setCouch(self.url, self.database)
        g1.create()

        g2 = Group(name = "DMWM")
        g2.setCouch(self.url, self.database)
        g2.get()

        g1.drop()

if __name__ == '__main__':
    unittest.main()



from __future__ import (print_function, division)

import unittest

from mock import mock

from WMCore.Storage.Backends.CPImpl import CPImpl


class CPImplTest(unittest.TestCase):
    def setUp(self):
        self.CPImpl = CPImpl()

    def testCreateSourceName_simple(self):
        self.assertEqual("name", self.CPImpl.createSourceName("", "name"))

    @mock.patch('WMCore.Storage.Backends.CPImpl.os')
    def testCreateOutputDirectory_noDir(self, mock_os):
        mock_os.path.dirname.return_value = "/dir1/dir2"
        mock_os.path.isdir.return_value = False
        self.CPImpl.createOutputDirectory("/dir1/dir2/file")
        mock_os.path.dirname.assert_called_once_with("/dir1/dir2/file")
        mock_os.path.isdir.assert_called_once_with("/dir1/dir2")
        mock_os.makedirs.assert_called_once_with("/dir1/dir2")

    @mock.patch('WMCore.Storage.Backends.CPImpl.os')
    def testCreateOutputDirectory_withDir(self, mock_os):
        mock_os.path.dirname.return_value = "/dir1/dir2"
        mock_os.path.isdir.return_value = True
        self.CPImpl.createOutputDirectory("/dir1/dir2/file")
        mock_os.path.dirname.assert_called_once_with("/dir1/dir2/file")
        mock_os.path.isdir.assert_called_once_with("/dir1/dir2")
        mock_os.makedirs.assert_not_called()

    def testCreateStageOutCommand_noStageInNoOptionsNoChecksum(self):
        self.CPImpl.stageIn = False
        results = self.CPImpl.createStageOutCommand("sourcePFN", "targetPFN")
        expectedResults = self.createStageOutCommandResults(False, "sourcePFN", "targetPFN")
        self.assertEqual(expectedResults, results)

    def testCreateStageOutCommand_stageInNoOptionsNoChecksum(self):
        self.CPImpl.stageIn = True
        results = self.CPImpl.createStageOutCommand("sourcePFN", "targetPFN")
        expectedResults = self.createStageOutCommandResults(True, "sourcePFN", "targetPFN")
        self.assertEqual(expectedResults, results)

    def createStageOutCommandResults(self, stageIn, sourcePFN, targetPFN):
        copyCommand = ""
        if stageIn:
            remotePFN, localPFN = sourcePFN, targetPFN
        else:
            remotePFN, localPFN = targetPFN, sourcePFN
            copyCommand += "LOCAL_SIZE=`stat -c%%s %s`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"
        copyCommand += "cp %s %s\n" % (sourcePFN, targetPFN)
        if stageIn:
            copyCommand += "LOCAL_SIZE=`stat -c%%s %s`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"
            removeCommand = ""
        else:
            removeCommand = "rm %s;" % remotePFN
        copyCommand += "REMOTE_SIZE=`stat -c%%s %s`\n" % remotePFN
        copyCommand += "echo \"Remote File Size is: $REMOTE_SIZE\"\n"
        copyCommand += "if [ $REMOTE_SIZE ] && [ $LOCAL_SIZE == $REMOTE_SIZE ]; then exit 0; "
        copyCommand += "else echo \"ERROR: Size Mismatch between local and SE\"; %s exit 60311 ; fi" % removeCommand
        return copyCommand

    @mock.patch('WMCore.Storage.Backends.CPImpl.os')
    def testRemoveFile(self, mock_os):
        self.CPImpl.removeFile("file")
        mock_os.remove.assert_called_once_with("file")



#!/usr/bin/env python
"""
    WorkQueue.Policy.End.SingleShot tests
"""




import unittest
import math
from functools import partial
from WMCore.WorkQueue.Policy.End.SingleShot import SingleShot
from WMCore.WorkQueue.DataStructs.WorkQueueElement import WorkQueueElement as WQE


class SingleShotTestCase(unittest.TestCase):

    def setUp(self):
        """Create workflow stuff"""
        self.policy = partial(SingleShot)

        # ones i made earlier
        self.parent = WQE(); self.parent.id = 1
        self.available = WQE(Status = 'Available', ParentQueueId = 1)
        self.acquired = WQE(Status = 'Acquired', ParentQueueId = 1)
        self.negotiating = WQE(Status = 'Negotiating', ParentQueueId = 1)
        self.done = WQE(Status = 'Done', PercentComplete = 100, PercentSuccess = 100, ParentQueueId = 1)
        self.failed = WQE(Status = 'Failed', PercentComplete = 100, PercentSuccess = 0, ParentQueueId = 1)

    def tearDown(self):
        pass


    def testSuccessThreshold(self):
        """Check threshold for success"""
        # no custom functionality left - just calls base class
        pass


if __name__ == '__main__':
    unittest.main()



'''
Created on Aug 10, 2010

@author: meloam
'''
import unittest
import shutil
from WMQuality.TestInit import TestInit

from WMCore.WMSpec.WMStep import WMStep
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.DataStructs.Job import Job

from WMCore.WMSpec.Steps.Templates.DeleteFiles import DeleteFiles as DeleteTemplate
from WMCore.WMSpec.Steps.Executors.DeleteFiles import DeleteFiles as DeleteExecutor
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
import WMCore.WMSpec.Steps.StepFactory as StepFactory

import os.path
import sys
import inspect
from nose.plugins.attrib import attr

import WMCore_t.WMSpec_t.Steps_t as ModuleLocator

class deleteFileTest(unittest.TestCase):


    def setUp(self):
        # stolen from CMSSWExecutor_t. thanks, dave
        self.testInit = TestInit(__file__)
        self.testDir = self.testInit.generateWorkDir()
        shutil.copyfile('/etc/hosts', os.path.join(self.testDir, 'testfile'))

        self.workload = newWorkload("UnitTests")
        self.task = self.workload.newTask("DeleterTask")
        stepHelper = step = self.task.makeStep("DeleteTest")
        self.step = stepHelper.data
        self.actualStep = stepHelper
        template = DeleteTemplate()
        template(self.step)
        self.helper = template.helper(self.step)
        self.executor = StepFactory.getStepExecutor(self.actualStep.stepType())

        taskMaker = TaskMaker(self.workload, self.testDir)
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        self.sandboxDir = "%s/UnitTests" % self.testDir

        self.task.build(self.testDir)
        sys.path.insert(0, self.testDir)
        sys.path.insert(0, self.sandboxDir)


        self.job = Job(name = "/UnitTest/DeleterTask/DeleteTest-test-job")

        binDir = inspect.getsourcefile(ModuleLocator)
        binDir = binDir.replace("__init__.py", "bin")

        if not binDir in os.environ['PATH']:
            os.environ['PATH'] = "%s:%s" % (os.environ['PATH'], binDir)

    def tearDown(self):
        self.testInit.delWorkDir()
        sys.path.remove(self.testDir)
        sys.path.remove(self.sandboxDir)

    def setLocalOverride(self, step):
        step.section_('override')
        step.override.command    = 'cp'
        step.override.option     = ''
        step.override.__setattr__('lfn-prefix', '')
        step.override.__setattr__('phedex-node','DUMMYPNN')


    @attr('integration')
    def testManualDeleteOld(self):
        self.assertTrue(os.path.exists( os.path.join(self.testDir, 'testfile')))
        self.step.section_('filesToDelete')
        self.step.filesToDelete.file1 = os.path.join(self.testDir, 'testfile')
        self.setLocalOverride(self.step)
        self.executor.initialise(self.step, self.job)
        self.executor.execute()
        self.assertFalse(os.path.exists( os.path.join(self.testDir, 'testfile')))
        return

    @attr('integration')
    def testManualDeleteNew(self):
        self.assertTrue(os.path.exists( os.path.join(self.testDir, 'testfile')))
        self.step.section_('filesToDelete')
        self.step.filesToDelete.file1 = os.path.join(self.testDir, 'testfile')
        self.setLocalOverride(self.step)
        self.step.override.newStageOut = True
        self.executor.initialise(self.step, self.job)
        self.executor.execute()
        self.assertFalse(os.path.exists( os.path.join(self.testDir, 'testfile')))
        return

    @attr('integration')
    def testJobDeleteOld(self):
        self.assertTrue(os.path.exists( os.path.join(self.testDir, 'testfile')))
        self.setLocalOverride(self.step)
        self.job['input_files'] = [ {'lfn': os.path.join(self.testDir, 'testfile') } ]
        self.executor.initialise(self.step, self.job)
        self.executor.execute()
        self.assertFalse(os.path.exists( os.path.join(self.testDir, 'testfile')))
        return

    @attr('integration')
    def testJobDeleteNew(self):
        self.assertTrue(os.path.exists( os.path.join(self.testDir, 'testfile')))
        self.setLocalOverride(self.step)
        self.step.override.newStageOut = True
        self.job['input_files'] = [ {'lfn': os.path.join(self.testDir, 'testfile') } ]
        self.executor.initialise(self.step, self.job)
        self.executor.execute()
        self.assertFalse(os.path.exists( os.path.join(self.testDir, 'testfile')))
        return

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
MakeDBSMockFiles

Program to create mock DBS JSON files used by the DBS mock-based emulator

If there are either dataset or dbs calls removal, then you need to create
new files from scratch (i.e., you need to delete the previous DBS json
mocked data).
"""

from __future__ import (division, print_function)

from future.utils import viewitems

import json
import os

from RestClient.ErrorHandling.RestClientExceptions import HTTPError

from WMCore.Services.DBS.DBSReader import DBSReader as DBSReader
from WMCore.WMBase import getTestBase
from WMQuality.Emulators.DBSClient.MockedDBSGlobalCalls import calls as calls00
from WMQuality.Emulators.DBSClient.MockedDBSGlobalCalls import datasets as datasets00
from WMQuality.Emulators.DBSClient.MockedDBSGlobalCalls import endpoint as endpoint00
from WMQuality.Emulators.DBSClient.MockedDBSphys03Calls import calls as calls03
from WMQuality.Emulators.DBSClient.MockedDBSphys03Calls import datasets as datasets03
from WMQuality.Emulators.DBSClient.MockedDBSphys03Calls import endpoint as endpoint03

INSTANCES = ((endpoint00, 'DBSMockData.json', calls00, datasets00),
             (endpoint03, 'DBSMockData03.json', calls03, datasets03))

for endpoint, outFile, calls, datasets in INSTANCES:
    lookup = {}
    outFilename = os.path.join(getTestBase(), '..', 'data', 'Mock', outFile)
    ## Instead of updating on top of the previous file, just go with a new one
    #try:
    #    with open(outFilename, 'r') as mockData:
    #        lookup = json.load(mockData)
    #except IOError:
    #    pass

    dbs = DBSReader(endpoint)
    realDBS = dbs.dbs

    for dataset in datasets:
        print("Building call list for", dataset)
        calls.append(['listDatasetParents', {'dataset': dataset}])
        calls.append(['listDatasets', {'dataset_access_type': '*', 'dataset': dataset}])
        calls.append(['listBlockOrigin', {'dataset': dataset}])
        calls.append(['listBlocks', {'dataset': dataset}])
        calls.append(['listBlocks', {'detail': False, 'dataset': dataset}])
        calls.append(['listBlocks', {'detail': True, 'dataset': dataset}])
        calls.append(['listFileSummaries', {'validFileOnly': 1, 'dataset': dataset}])
        calls.append(['listFileArray', {'dataset': dataset, 'detail': True, 'validFileOnly': 1}])
        calls.append(['listFileArray', {'dataset': dataset, 'detail': False, 'validFileOnly': 1}])
        calls.append(['listRuns', {'dataset': dataset}])
        blocks = realDBS.listBlocks(dataset=dataset)
        for block in blocks:
            calls.append(['listBlocks', {'block_name': block['block_name']}])
            calls.append(['listBlocks', {'block_name': block['block_name'], 'detail': True}])
            calls.append(['listBlockParents', {'block_name': str(block['block_name'])}])
            calls.append(['listFileLumis', {'block_name': block['block_name']}])
            calls.append(['listFileLumis', {'block_name': block['block_name'], 'validFileOnly': 1}])
            calls.append(['listFileArray', {'block_name': block['block_name'],
                                            'detail': True, 'validFileOnly': 1}])
            calls.append(['listFileSummaries', {'block_name': block['block_name'], 'validFileOnly': 1}])
            calls.append(['listRuns', {'block_name': block['block_name']}])
            calls.append(['listFileParents', {'block_name': block['block_name']}])
            files = realDBS.listFiles(block_name=block['block_name'])
            for dbsFile in files:
                lfn = dbsFile['logical_file_name']
                calls.append(['listFileArray', {'logical_file_name': [lfn], 'detail': True}])
                calls.append(['listFileArray', {'logical_file_name': [lfn]}])
                calls.append(['listFileLumiArray', {'logical_file_name': [lfn]}])

    nCalls = len(calls)
    print("Need to issue %d calls to DBS" % nCalls)
    callsDone = 0
    progress = [2.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.05, 0.03, 0.02, 0.01]
    for call in calls:
        callsDone += 1
        if callsDone / nCalls > progress[-1]:
            percentDone = progress.pop() * 100
            print(" Fetching call list %d%% done." % percentDone)
        func = getattr(realDBS, call[0])
        if len(call) > 1:
            signature = '%s:%s' % (call[0], sorted(viewitems(call[1])))
            try:
                result = func(**call[1])
            except HTTPError:
                result = 'Raises HTTPError'
        else:
            result = func()
            signature = call[0]

        lookup.update({signature: result})

    print("Writing out %s file" % outFilename)
    with open(outFilename, 'w') as mockData:
        json.dump(lookup, mockData, indent=1, separators=(',', ': '), sort_keys=True)



#!/usr/bin/env python

"""
MySQL implementation of AddChecksum
"""





from WMCore.Database.DBFormatter import DBFormatter

class AddChecksum(DBFormatter):
    sql = """INSERT INTO dbsbuffer_file_checksums (fileid, typeid, cksum)
             SELECT :fileid, (SELECT id FROM dbsbuffer_checksum_type WHERE type = :cktype), :cksum FROM dual
             WHERE NOT EXISTS (SELECT fileid FROM dbsbuffer_file_checksums WHERE
                               fileid = :fileid AND typeid = (SELECT id FROM dbsbuffer_checksum_type WHERE type = :cktype))"""

    def execute(self, fileid = None, cktype = None, cksum = None, bulkList = None, conn = None,
                transaction = False):

        if bulkList:
            binds = bulkList
        else:
            binds = {'fileid': fileid, 'cktype': cktype, 'cksum': cksum}

        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Oracle implementation for retrieving the total number of jobs
for a given workflow in a given wmbs status
"""

from __future__ import print_function, division

from WMCore.WMBS.MySQL.Jobs.GetNumberOfJobsForWorkflowStatus import GetNumberOfJobsForWorkflowStatus \
    as MySQLGetNumberOfJobsForWorkflowStatus


class GetNumberOfJobsForWorkflowStatus(MySQLGetNumberOfJobsForWorkflowStatus):
    """
    Exactly the same implementation as of the MySQL one
    """
    pass



from __future__ import print_function, division

from builtins import object
from future.utils import viewitems

from WMCore.ReqMgr.DataStructs.DefaultConfig.EDITABLE_SPLITTING_PARAM_CONFIG import EDITABLE_SPLITTING_PARAM_CONFIG
from WMCore.ReqMgr.DataStructs.DefaultConfig.DAS_RESULT_FILTER import DAS_RESULT_FILTER

#TODO need to find the better way to import all the config from DefaultConfig area
DEFAULT_CONFIG = {'EDITABLE_SPLITTING_PARAM_CONFIG': EDITABLE_SPLITTING_PARAM_CONFIG,
                  'DAS_RESULT_FILTER': DAS_RESULT_FILTER}

class ReqMgrConfigDataCache(object):
    # this cache shouldn't be accessed directly
    # Only through the reqmgr2 api
    _req_aux_db = None
    _req_config_data_cache = {}

    @staticmethod
    def set_aux_db(couchdb):
        ReqMgrConfigDataCache._req_aux_db = couchdb
        return

    @staticmethod
    def register_default(default_config):
        ReqMgrConfigDataCache._req_config_data_cache.update(default_config)
        return

    @staticmethod
    def getConfig(doc_name):

        try:
            config = ReqMgrConfigDataCache._req_aux_db.document(doc_name)
            del config["_id"]
            del config["_rev"]
        # TODO only get the exception when server is not available.
        except Exception as ex:
            config = ReqMgrConfigDataCache._req_config_data_cache.get(doc_name, None)
            print(f"Getting document '{doc_name}' from cache due to error: {str(ex)}")
        return config

    @staticmethod
    def replaceConfig(doc_name, content):
        if doc_name not in ReqMgrConfigDataCache._req_config_data_cache:
            raise Exception("No config exists")

        if ReqMgrConfigDataCache._req_aux_db.documentExists(doc_name):
            ReqMgrConfigDataCache._req_aux_db.delete_doc(doc_name)
        response = ReqMgrConfigDataCache._req_aux_db.putDocument(doc_name, content)
        ReqMgrConfigDataCache._req_config_data_cache.update(doc_name = content)
        return response

    @staticmethod
    def putDefaultConfig():
        error = ""
        for doc_name, content in viewitems(DEFAULT_CONFIG):
            try:
                ReqMgrConfigDataCache.replaceConfig(doc_name, content)
            except Exception as ex:
                error += str(ex)

        return error

# register all default config
ReqMgrConfigDataCache.register_default(DEFAULT_CONFIG)



#!/usr/bin/env python
"""
_LoadJobs_

Oracle implementation of JobGroup.LoadJobs
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.LoadJobs import LoadJobs as LoadJobsJobGroupMySQL

class LoadJobs(LoadJobsJobGroupMySQL):
    sql = LoadJobsJobGroupMySQL.sql



#!/usr/bin/env python
"""
_ListWorkflow_

Oracle implementation of DBS3Buffer.ListWorkflow

Created on May 2, 2013

@author: dballest
"""

from WMComponent.DBS3Buffer.MySQL.ListWorkflow import ListWorkflow as MySQLListWorkflow

class ListWorkflow(MySQLListWorkflow):
    pass
